'''
Run all water plant lops between demand side coils and supply side boilers/chillers
'''
from eagers.building.process.flow_resolver_demand import flow_resolver_demand
from eagers.building.process.demand_loop import demand_loop
from eagers.building.process.flow_resolver_supply import flow_resolver_supply
from eagers.building.process.supply_loop import supply_loop
from eagers.building.components.half_loop_tank import half_loop_tank

def manage_plant_equip(building,profile,plant_nodes,e_use,dt,T_air,w_air,T_mains,wh_tank,T_supply0,T_return0):   
    ##This function runs the plant water loop supply and demand side 
    ##The correct order matters so that a supply loop (chillers) puts load on a demand loop (cooling tower/condenser)
    ##This function determines 'ideal' heating and cooling and simulates supply
    ##equipment settings (PLR) to provide this amount of heating/cooling
    cp_water = 4186  #J/kg*K
    nl = len(building.plant_loop['name'])
    loop_load  = [[0 for i in range(nl)] for j in range(3)]
    T_supply_avg = [j for j in T_supply0]
    for i  in range(nl):
        l = building.plant_loop['loop_order'].index(i)
        n_d_in = building.plant_demand_nodes['inlet_node'][l]
        n_d_out = building.plant_demand_nodes['outlet_node'][l]
        n_s_in = building.plant_supply_nodes['inlet_node'][l]
        n_s_out = building.plant_supply_nodes['outlet_node'][l]
        sup_lv = building.plant_loop['loop_volume'][l]/2
        on_loop = [i for i,x in enumerate(building.plant_supply_equip['loop']) if x==l]
        for j in on_loop:
            if building.plant_supply_equip['type'][j]== 'waterheater:mixed':
                sup_lv += building.water_heater[building.plant_supply_equip['name'][j]]['tank_volume']

        ## Plant demand loops
        plant_nodes,pump_power,pump_heat = flow_resolver_demand(building,plant_nodes,profile,l,T_mains) 
        plant_nodes = demand_loop(building,profile,T_mains,plant_nodes,l) 
        loop_flow_rate = plant_nodes['demand_flow'][n_d_out] 
        loop_return_T = plant_nodes['demand_temperature'][n_d_out]
        T_return,T_return_avg = half_loop_tank(building.plant_loop['loop_volume'][l]/2,building.plant_loop['fluid_density'][l],loop_flow_rate,T_return0[l],pump_heat,loop_return_T,dt) # connect demand half loop to supply half loop
        plant_nodes['supply_temperature'][n_s_in] = T_return_avg+0
        plant_nodes['supply_flow'][n_s_in] = loop_flow_rate+0

        ## loop load (including tank back to nominal setpoint)
        if building.plant_supply_nodes['name'][n_s_out] in building.manager['node']:
            m = building.manager['node'].index(building.plant_supply_nodes['name'][n_s_out])
            if building.manager['type'][m] == 'Scheduled':
                T_treat = profile[building.manager['schedule'][m]]
            else:
                T_treat = building.plant_loop['exit_temperature'][l]
        else:
            T_treat = building.plant_loop['exit_temperature'][l]

        if building.plant_loop['type'][l].lower()=='heating':
            if loop_flow_rate>0:
                tank_heat = (T_treat - T_supply_avg[l])*cp_water*building.plant_loop['loop_volume'][l]/2*building.plant_loop['fluid_density'][l]#energy in J
            else: 
                tank_heat = 0
            loop_load[0][l] = max(0,loop_flow_rate*cp_water*(T_treat-T_return_avg) - pump_heat + tank_heat/dt)#power % (Watts) kg/s*J/kgK * K = J/s
            supply_load = loop_load[0][l]
        else:
            if loop_flow_rate>0:
                tank_cool = (T_supply_avg[l] - T_treat)*cp_water*building.plant_loop['loop_volume'][l]/2*building.plant_loop['fluid_density'][l]#energy in J
            else: 
                tank_cool = 0
            if building.plant_loop['type'][l].lower()=='cooling':
                loop_load[1][l] = max(0,loop_flow_rate*cp_water*(T_return_avg-T_treat) + pump_heat + tank_cool/dt)#power (Watts) kg/s*J/kgK * K = J/s
                supply_load = loop_load[1][l]
            elif building.plant_loop['type'][l].lower()=='condenser':
                loop_load[2][l] = max(0,loop_flow_rate*cp_water*(T_return_avg-T_treat) + pump_heat + tank_cool/dt)#power % (Watts) kg/s*J/kgK * K = J/s
                supply_load = loop_load[2][l]

        ## Plant supply loops        
        plant_nodes['supply_flow'],plant_nodes['supply_temperature'] = flow_resolver_supply(building,plant_nodes['supply_temperature'],supply_load,loop_flow_rate,T_return_avg,building.plant_loop['load_scheme'][l],on_loop) 
        plant_nodes,wh_tank,cool_elec,heat_gas,water_gas,water_elec,tower_elec = supply_loop(building,profile,plant_nodes,wh_tank,T_air,w_air,l,dt) 
        treated_T = plant_nodes['supply_temperature'][n_s_out]
        T_supply_new,T_avg = half_loop_tank(sup_lv,building.plant_loop['fluid_density'][l],plant_nodes['supply_flow'][n_s_out],T_supply0[l],0,treated_T,dt) # connect supply half loop to demand half loop  
        T_supply_avg[l] = .15*T_supply_avg[l] + .85*T_avg
        plant_nodes['demand_temperature'][n_d_in] = T_supply_avg[l]+0

        e_use['pumps'] += pump_power #Watts
        e_use['cool_elec'] += cool_elec #Watts
        e_use['heat_gas'] += heat_gas #Watts
        e_use['water_gas'] +=water_gas
        e_use['water_elec'] += water_elec
        e_use['tower_elec'] += tower_elec#Watts
        plant_nodes['demand_temperature'][n_d_in] = T_avg+0
    return plant_nodes,e_use,wh_tank,loop_load


"""Defines class DataConstructed, which provides a versatile constructor
method: from_data().

Classes:
DataConstructed
"""



class DataConstructed:
    @classmethod
    def from_data(cls, data, as_list=False):
        """Returns an instance from a given instance, dict, str, list,
        or NoneType. If optional argument as_list is specified as True,
        the returned value will be a list of class instances or an empty
        list in the case of a NoneType.
        """
        if isinstance(data, cls):
            instance = data
        elif isinstance(data, dict):
            # Handle special field _type if present.
            try:
                subcls_name = data.pop('_type')
            except KeyError:
                subcls_name = None
            subcls = (
                cls if subcls_name is None
                else cls._class_from_name(subcls_name)
            )
            return subcls(**data)
        elif isinstance(data, str):
                instance = cls.from_file(data)
        elif isinstance(data, list):
            return [cls.from_data(x) for x in data]
        elif data is None:
            if as_list:
                return []
            else:
                return cls()
        else:
            raise TypeError(f"Unexpected type {type(data)!r}")
        # Ensure return type of list if that was specified.
        if as_list:
            return [instance]
        return instance



from eagers.basic.hdf5 import DatetimeFloatConverter as DFC

def auto_set_startdate(test_data):
    """Automatically set the value of start_date based on the given
    TestData. If start_date is already set, this does nothing.

    Positional arguments:
    test_data - (TestData) TestData instance used to find start
        date.
    """
    if 'nodedata' in test_data:
        tdfield = 'nodedata'
    elif 'hydro' in test_data:
        tdfield = 'hydro'
    else:
        tdfield = 'weather'
    # Use the first date common to all tables with a 'timestamp'
    # header as the start date.
    data = test_data[tdfield]
    timestamp_tables = [t for t in data.tablenames if 'timestamp' in data.colnames(t)]
    first_dates = (test_data[tdfield].read(t, 0, 1, field='timestamp')[0] for t in timestamp_tables)
    sd = DFC.f2d_sgl2sgl(max(first_dates))
    return sd


from copy import copy
from eagers.building.components.fan_calc import fan_calc
from eagers.building.basic.psychometric import psychometric
from eagers.building.components.cooling_coil import cooling_coil
from eagers.building.components.water_coil import water_coil

def unitary_systems(building,air_nodes,supply_flow,direct_flow,direct_return_air,T_direct,T_air,w_air,fan_avail,plant_nodes,e_use):
    ## Unitary systems (zones not on an air loop)
    r_air = {}
    c_air = {}
    air_density = 1.204
    for i in building.unitary_sys:
        z = building.unitary_sys[i]['zone']
        if z>=0 and direct_flow[z]>0:
            m = building.zones['multiplier'][z]
            fan_name = building.unitary_sys[i]['fan_name']
            for x in direct_return_air:
                r_air[x] = direct_return_air[x][z]
            c_air['Tdb'] = supply_flow['Tdb'][z]
            c_air['w'] = supply_flow['w'][z] 
            c_air['h'] = supply_flow['h'][z]
            c_air['m_dot'] = supply_flow['m_dot'][z]*m
            if 'terminal' in building.unitary_sys[i]:
                r_air['m_dot'] = direct_flow[z]*m*air_density - c_air['m_dot'] 
                air = mix_flows(r_air,c_air) #fresh air directly to zone, not outdoor air mixing on loop
            else:
                r_air['m_dot'] = direct_return_air['m_dot'][z]*m 
                fresh_air = {}
                fresh_air['Tdb'] = T_air
                fresh_air['w'] = w_air
                fresh_air['h'] = psychometric(fresh_air,'h')
                fresh_air['m_dot'] = direct_flow[z]*m*air_density - r_air['m_dot']
                air = mix_flows(r_air,fresh_air) #fresh air directly to zone, not outdoor air mixing on loop
            air,P_fan,_ =  fan_calc(building.fans[fan_name],air,fan_avail[fan_name])
            if ('end_use' in building.fans[fan_name] and building.fans[fan_name]['end_use'] == 'fan energy'):
                e_use['fan_elec'] += P_fan
            else: # 'Unit Heater Fans', 'Zone Exhaust Fans'c
                e_use['other_fans'] += P_fan

            if building.unitary_sys[i]['type'] == 'FourPipeFanCoil':
                air,plant_nodes,e_use = unit_cool_coil(building,building.unitary_sys[i]['cool_coil'],building.unitary_sys[i]['cool_coil_type'],air,T_direct[z],T_air,e_use,plant_nodes,i)
                air,plant_nodes,e_use = unit_heat_coil(building,building.unitary_sys[i]['heat_coil'],building.unitary_sys[i]['heat_coil_type'],plant_nodes,air,T_direct[z],e_use,i)
            elif building.unitary_sys[i]['type'] == 'PackagedTerminalAirConditioner':
                air,plant_nodes,e_use = unit_cool_coil(building,building.unitary_sys[i]['cool_coil'],building.unitary_sys[i]['cool_coil_type'],air,T_direct[z],T_air,e_use,plant_nodes,i)
                air,plant_nodes,e_use = unit_heat_coil(building,building.unitary_sys[i]['heat_coil'],building.unitary_sys[i]['heat_coil_type'],plant_nodes,air,T_direct[z],e_use,i)
            elif building.unitary_sys[i]['type']== 'UnitHeater':
                air,plant_nodes,e_use = unit_heat_coil(building,building.unitary_sys[i]['heat_coil'],building.unitary_sys[i]['heat_coil_type'],plant_nodes,air,T_direct[z],e_use,i)
            elif building.unitary_sys[i]['type'] =='WindowAirConditioner':
                    pass#not sure what to do about this (in zone, but not on air loop?)
            supply_flow['Tdb'][z] = air['Tdb']
            supply_flow['w'][z] = air['w']
            supply_flow['h'][z] = air['h']
            supply_flow['m_dot'][z] = air['m_dot']
    return supply_flow,plant_nodes,e_use

    
def unit_cool_coil(building,coil_name,coil_type,air,T_supply,T_air,e_use,plant_nodes,i):
    if (coil_type == 'coil:cooling:dx:twospeed' or coil_type == 'coil:cooling:dx:singlespeed'):
        air,P_coil = cooling_coil(building,coil_name,air,T_air,T_supply)
        e_use['cool_elec'] += P_coil
    elif (coil_type == 'coil:cooling:water:detailedgeometry' or coil_type == 'coil:cooling:water'):
        air,m_w,actual_load,out = water_coil(building,coil_name,air,T_supply,plant_nodes)
        plant_nodes['demand_flow'][out] = m_w #kg/s
        plant_nodes['load'][out] = actual_load #load in Watts (negative because air leaves hotter)
    return air,plant_nodes,e_use

def unit_heat_coil(building,coil_name,coil_type,plant_nodes,air,T_supply,e_use,i):
    if (coil_type == 'coil:heating:fuel' or coil_type == 'coil:heating:electric'):
        cp_supply = 1006 + 1860*air['w']
        load_heat = max(0,air['m_dot']*cp_supply*(T_supply-air['Tdb']))
        if load_heat>0:
            air['Tdb'] += load_heat/(cp_supply*air['m_dot'])
            air['h'] = psychometric(air,'h')
            if coil_type == 'coil:heating:fuel':
                e_use['heat_gas'] += load_heat/building.coils_heating[coil_name]['efficiency'] #heat added in W
            else:
                e_use['heat_elec'] += load_heat/building.coils_heating[coil_name]['efficiency'] #heat added in W
    elif coil_type == 'coil:heating:water':
        air,m_w,actual_load,out = water_coil(building,coil_name,air,T_supply,plant_nodes)
        plant_nodes['demand_flow'][out] = m_w #kg/s
        plant_nodes['load'][out] = actual_load #load in Watts (negative because air leaves hotter)
    return air,plant_nodes,e_use

def mix_flows(s1,s2):
    mixed = {}
    mixed['m_dot'] = s1['m_dot'] + s2['m_dot'] #mass flow (kg/s) :
    mixed['h'] = (s1['m_dot']*s1['h'] + s2['m_dot']*s2['h'])/mixed['m_dot']
    w_sat = psychometric({'h':mixed['h']},'w')
    mix_w = (s1['m_dot']*s1['w'] + s2['m_dot']*s2['w'])/mixed['m_dot']
    mixed['m_dot'] = mixed['m_dot']*(1 - max(0,mix_w-w_sat))
    mixed['w'] = min(w_sat,mix_w)
    mixed['Tdb'] = psychometric(mixed,'Tdb')
    return mixed



"""Functionality for reading data sets from JSON files.

Defines:
read_dataset
"""

import json

from dateutil import parser

from eagers.config.path_spec import JSON_SUFFIX, USER_DIR_DATASETS


def read_dataset(dataset_class, name, convert_timestamp=True):
    """Read a data set from a JSON file to a dictionary.

    Positional arguments:
    dataset_class - (str) Data set class name.
    name - (str) Name of the JSON file.

    Keyword arguments:
    convert_timestamp - (bool) When True, the resulting dictionary will
        be checked for a key called 'timestamp'. If it exists, all its
        values will be converted to datetimes.
    """
    # Ensure ".json" suffix.
    try:
        suffix = name.split('.')[1]
        if suffix != 'json':
            raise ValueError(
                f'File name {name} has a non-JSON suffix.')
    except IndexError:
        # No file suffix; add one.
        name += JSON_SUFFIX
    # Get file path.
    filepath = USER_DIR_DATASETS / dataset_class / name
    # Read.
    with open(filepath, 'r') as jf:
        result = json.load(jf)
    if convert_timestamp and 'timestamp' in result:
        result['timestamp'] = convert_to_datetime(result['timestamp'])
    return result


def convert_to_datetime(t):
    """Recursively convert the given string or sequence to datetime(s).
    If the given object is of neither type, return the object.
    """
    if isinstance(t, str):
        return parser.parse(t)
    elif isinstance(t, (list, tuple, set)):
        seq_type = type(t)
        return seq_type(map(convert_to_datetime, t))
    else:
        return t



from eagers.building.basic.psychometric import psychometric

import numpy as np

def cooling_tower(building,name,water_flow,T_in,T_target,T_air,w_air):
    tower = building.cooling_tower[name] 
    bypass = 0
    if tower['type']=='SingleSpeed':
        if water_flow>0:
            T_out_fc = tower_iteration(tower['free_convect_UA'],water_flow,tower['free_convect_flow'],T_in,T_air,w_air,T_target)
            T_out_fan = tower_iteration(tower['UA'],water_flow,tower['air_flow'],T_in,T_air,w_air,T_target)       
            if tower['capacity_control']=='fancycling':
                if T_out_fan == T_out_fc:
                    frac_fan = 0
                else:
                    frac_fan = max(0,min(1,(T_target-max(T_target,T_out_fc))/(T_out_fan - T_out_fc)))
                if frac_fan == 0:
                    bypass = 1- (T_in - T_target)/(T_in - T_out_fc)
                    P_fan = 0
                    T_out = T_target
                else:
                    P_fan = frac_fan*tower['fan_power']#Watts
                    T_out = T_out_fc - frac_fan*(T_out_fc - T_out_fan)
            else:
                P_fan = tower['fan_power']#Watts
                T_out = T_out_fan
        else:
            P_fan = 0
            T_out = T_in
            bypass = 0
    else:
        print('need cooling tower model for type__'+tower['type'])
    return P_fan,T_out,bypass

def tower_iteration(UA,water_flow,air_flow,T_in,T_air,w_air,T_wb_exit):
    Cp_water = 4186 #J/kg*K
    c_w = water_flow*Cp_water
    T_wb = psychometric({'Tdb':T_air,'w':w_air},'Twb')
    T_in = max(T_in, T_wb+4)
    T_wb_exit = max(T_wb_exit,T_wb+1)
    T0 = T_wb_exit
    ##iterate the following
    error_T = 1
    d_T = 0.1
    gain =.5
    count = 1
    while abs(error_T)>1e-2:
        Q_total,c_a = tower_heat_trans(T_in,T_air,w_air,T_wb,T_wb_exit,air_flow,c_w,UA)
        Q_total2,_ = tower_heat_trans(T_in,T_air,w_air,T_wb,T_wb_exit+d_T,air_flow,c_w,UA)
        if Q_total == Q_total2:
            T_out = T_wb_exit
            break
        Q_tran = c_a*(T_wb_exit-T_wb)#heat transfer to get to this exit temperature
        dQ_dT = (Q_total2-Q_total)/d_T#marginal heat transfer per degree Twb_exit
        error_T = (Q_tran - Q_total)/dQ_dT
        T_wb_exit = T_wb_exit - gain*error_T
        if T_wb_exit>T_in  or T_wb_exit<T_wb:
            T_wb_exit = T0
        if count>10:
            gain = max(.03125,gain*.5)
            if d_T<1e-2:
                error_T = 0
            d_T = d_T*.25
            count = 0
        count = count+1
    T_out = T_in - Q_total/c_w
    return T_out

def tower_heat_trans(T_in,T_air,w_air,T_wb,T_wb_exit,air_flow,c_w,UA):
    air_density = 1.204 #kg/m^3 #standard air density (1.204 kg/m3) adjusted for the local barometric pressure (standard barometric pressure corrected for altitude, ASHRAE 1997 HOF pg. 6.1).
    Cp_air = 1034 #J/kg*K
    # Cp_e = 2*Cp_air
    Cp_e = (psychometric({'Twb':T_wb_exit,'w':w_air},'h') - psychometric({'Tdb':T_air,'w':w_air},'h'))/(T_wb_exit - T_wb) #J/kg*K
    c_a = air_flow*air_density*Cp_e
    UA_e = UA*Cp_e/Cp_air
    c_min = min(c_w,c_a)
    c_max = max(c_w,c_a)
    NTU = UA_e/c_min
    a = -NTU*(1-c_min/c_max)
    e = float((1-np.exp(a))/(1 - c_min/c_max*np.exp(a)))
    Q_total = e*c_min*(T_in - T_wb)
    return Q_total,c_a


def read_network(network):
    ''' interpret network  information'''
    for net_i in network:
        if not 'buildings' in net_i:
            net_i['buildings'] = None
        net_i['location'] = {}
        net_i['location']['latitude'] = net_i['latitude']
        net_i['location']['longitude'] = net_i['longitude']
        net_i['location']['time_zone'] = net_i['time_zone']
        if 'connections' in net_i:
            data = net_i['connections']
            test_con = True
            headers = list(data.keys())
            j = 1
            con = []
            limit = []
            eff = []
            while test_con:
                n = 'Connexn' + str(j) + ':node_name'
                if n in headers:
                    l = 'Connexn' + str(j) + ':line_limit'
                    e = 'Connexn' + str(j) + ':line_efficiency'
                    con.append(data[n])
                    limit.append(data[l])
                    eff.append(data[e])
                    del data[n]
                    del data[l]
                    del data[e]
                    j += 1
                else:
                    test_con = False
            if len(con) > 0:
                data['connections'] = {'node_name': con, 'line_limit': limit, 'line_efficiency': eff}
        else:
            net_i['connections'] ={}
            net_i['connections']['node_name'] = None
            net_i['connections']['line_limit'] = None
            net_i['connections']['line_efficiency'] = None


from eagers.building.plot_sim import plot_sim
from eagers.building.process.building_warmup import building_warmup
from eagers.building.process.run_building import run_building
from eagers.building.process.run_sim import run_sim
from eagers.building.basic.interpolate_weather import interpolate_weather
from eagers.class_definition.building import Building
from eagers.config.building import BUILD_NAMES, CLIMATE_NAMES
from eagers.config.path_spec import USER_DIR_BUILDINGS, USER_DIR_ENERGYPLUS_WEATHER, USER_DIR_EPLUS, DEMO_DIR
from eagers.read.weather_file import read_weather_file
from eagers.read.load_data import load_data
import pickle

def validate_building(build = 1,climate = '4A', start_day = 1, duration = 10, horizon = 0, mode = 'build',warmup = 'new'):
    ''''sim' for model predictive control simulation, 'build' for simulating a building without external control, 'warmup' just to warm up and save a model
      #'new' for new initialization, 'old' for a saved initialization

      buildings 2, 3, 4, 7, 8, 10 have flexibility (variable speed fans)
      build = input('Which building to run, name or 1-16?:')
      start_day = input('What day of the year will you start from (1-365)?:') 
      duaration = input('How many days will you simulate?:')
    '''
    if warmup == 'new':
        if type(build) is 'int':
            build = BUILD_NAMES[build-1]
        climate = CLIMATE_NAMES[climate]
        try:
            weather = read_weather_file(DEMO_DIR / (climate + '_TMY2.epw'),'tmy')
            building = Building.from_file(DEMO_DIR / ('RefBldg'+build + 'New2004_v1.4_7.2_' + climate))
        except:
            weather = read_weather_file(USER_DIR_ENERGYPLUS_WEATHER / (climate + '_TMY2.epw'),'tmy')
            building = Building.from_file(USER_DIR_BUILDINGS / ('RefBldg'+build + 'New2004_v1.4_7.2_' + climate))

        T_zone,T_surf,humidity,frost = building_warmup(building, weather, building.sim_date[24*(start_day-1)], False)
        
        date = [building.sim_date[i+24*(start_day-1)] for i in range(24*duration+1)]
        if len(weather['t_dryb']) != (len(date)+horizon):
            w_date = [building.sim_date[i+24*(start_day-1)] for i in range(24*duration+horizon)]
            weather = interpolate_weather(weather, w_date)
        if mode == 'warmup': # save warmup variables
            f_name = input('Name the saved building model:')
            with open(f_name+'.pckl', 'wb') as f:
                pickle.dump([T_zone, T_surf, humidity, weather, building], f)
    elif warmup == 'old':
        f_name = input('Specify name of the saved building model:')
        with open(f_name + '.pckl', 'rb') as f:
            T_zone, T_surf, humidity, weather, building = pickle.load(f)

    if mode == 'build':
        T_zone, T_surf, m_v_zone, frost, m_sys, Q_transfer, facility, Q_test_plots = run_building(building, [T_zone[-1]], [T_surf[-1]], [humidity[-1]], weather, date, False)
        try:
            plot_sim(facility,T_zone,Q_transfer,building,date[1:],DEMO_DIR / ('RefBldg'+build + 'New2004_v1.4_7.2_' + climate+'.csv'))
        except:
            plot_sim(facility,T_zone,Q_transfer,building,date[1:],USER_DIR_EPLUS / ('RefBldg'+build + 'New2004_v1.4_7.2_' + climate+'.csv'))
    elif mode == 'sim':
        test_data = load_data(building, weather, date, T_zone[-1], humidity[-1])##load test_data: actual internal gains and weather different from forecast schedules?
        facility, net_elec = run_sim(building, test_data, weather, [T_zone[-1]], [T_surf[-1]], [humidity[-1]], date, horizon)


'''Defines functionality for calculating generator limits.

Functions:
gen_limit
chp_heat
'''


def gen_limit(gen, gen_output, binary, dt, s):
    '''Return the generator limit.'''
    n_g = len(gen)
    n_s = len(dt)
    ramp_up = [[] for i in range(n_g)]
    ramp_down = [[] for i in range(n_g)]
    upper_bound = [0 for i in range(n_g)]
    lower_bound = [0 for i in range(n_g)]
    constraint = [[None for t in range(n_s)] for i in range(n_g)]

    acdc = False
    for i in range(n_g):
        if 'ramp' in gen[i] and not 'stor' in gen[i]:
            ramp_up[i] = [gen[i]['ramp']['b'][0] * dt[t] for t in range(n_s)]
            ramp_down[i] = [gen[i]['ramp']['b'][1] * dt[t] for t in range(n_s)]
            starts = [t for t in range(n_s) if binary[i][t + 1] and not binary[i][t]]
            stops = [t for t in range(n_s) if not binary[i][t + 1] and binary[i][t]]
            lower_bound[i] = gen[i][gen[i]['states'][0][0]]['lb'][-1]
            for st in starts:
                # If the ramp rate is less than the lb, increase ramp
                # rate at moment of startup.
                ramp_up[i][st] = max(ramp_up[i][st], lower_bound[i])
            for st in stops:
                # If the ramp rate is less than the lb, increase ramp
                # rate at moment of shutdown.
                ramp_down[i][st] = max(ramp_down[i][st], lower_bound[i])
            upper_bound[i] = gen[i]['ub']
        if gen[i]['type'] == 'ACDCConverter':
            acdc = True
            dc_ac = gen[i]['output']['e'][0][1]
            ac_dc = gen[i]['output']['dc'][0][0]
    if len(s) == 1 and (s[0] == 'e' or s[0] == 'dc') and any([g['type']== 'CombinedHeatPower' for g in gen]):
        # For combined heat and power cases.
        s.append('h')
    if acdc and not 'dc' in s:
        s.append('dc')
    if acdc and not 'e' in s:
        s.append('e')
    if 'h' in s:
        elec = False
        direct = False
        for i in range(n_g):
            if gen[i]['type'] == 'CombinedHeatPower':
                if 'e' in gen[i]['output']:
                    elec = True
                if 'dc' in gen[i]['output']:
                    direct = True
        if not 'e' in s and elec:
            s.append('e')
        if not 'dc' in s and direct:
            s.append('dc')
    s2 = {
        'e': ['CombinedHeatPower', 'ElectricGenerator'],
        'dc': ['CombinedHeatPower', 'ElectricGenerator'],
        'h': ['CombinedHeatPower','Heater'],
        'c': ['Chiller'],
        'cw': ['Chiller', 'CoolingTower'],
        'hy': ['Electrolyzer', 'HydrogenGenerator'],
    }
    spare_gen = {}
    max_out = {}
    inc = {}
    heat_output = [[] for i in range(n_g)]
    for k in s:
        spare_gen[k] = [0 for i in range(n_s)]
        max_out[k] = [[0 for t in range(n_s + 1)] for i in range(n_g)]
        inc[k] = []
        for i in range(n_g):
            if not 'stor' in gen[i] and gen[i]['type'] in s2[k]:
                if k == 'e' and acdc and 'dc' in gen[i]['output']:
                    inc[k].append(i)
                elif k == 'dc' and acdc and 'e' in gen[i]['output']:
                    inc[k].append(i)
                elif k in gen[i]['output']:
                    inc[k].append(i)
        for i in range(n_g):
            if i in inc[k]:
                max_out[k][i][0] = gen_output[i][0]
                # Constrained by initial condition (can't do anything).
                start = None
                for t in range(n_s):
                    if binary[i][t + 1]:
                        if not binary[i][t]:
                            start = t  # Just turned on.
                        if upper_bound[i] <= max_out[k][i][t] + ramp_up[i][t]:
                            max_out[k][i][t + 1] = upper_bound[i]
                        else:
                            max_out[k][i][t + 1] = max_out[k][i][t] + ramp_up[i][t]
                        if not start is None:
                            # Last index (0->n_s-1) when gen was off where it could be turned on early if neccesary to satisfy ramping constraint.
                            constraint[i][t] = start - 1
                for t in range(n_s - 1, 0, -1):
                    if (max_out[k][i][t] - ramp_down[i][t]) > max_out[k][i][t+1]:
                        # Constrained by shutdown.
                        max_out[k][i][t] = min([upper_bound[i],(max_out[k][i][t + 1] + ramp_down[i][t])])
                        if constraint[i][t - 1] is None:
                            # First index (0-->n_s-1) where gen has shut down, where it could be left on to satisfy ramping constraint.
                            constraint[i][t - 1] = min([z - 1 for z in range(t + 1, n_s + 1) if max_out[k][i][z] == 0])
                if k == 'h' and gen[i]['type'] == 'CombinedHeatPower':# Convert GenOutput to heatOutput.
                    heat_output[i] = chp_heat(gen[i], gen_output[i])
                    if 'e' in gen[i]['output']:
                        max_out['h'][i] = chp_heat(gen[i], max_out[k][i])
                    else:
                        max_out['h'][i] = chp_heat(gen[i], max_out[k][i])
                if gen[i]['type'] == 'Chiller' and gen[i]['source'] == 'heat':
                    # Assume net heat production remains the same, so do
                    # not count spare absorption chiller capacity.
                    # Don't count towards spare capacity.
                    max_out[k][i] = gen_output[i]
            if gen[i]['type'] == 'Utility' and k in gen[i]['output']:
                max_out[k][i] = [gen[i]['x']['ub']]
                # max_out[k][i].extend([gen_output[i][j+1] for j in range(n_s)])
                max_out[k][i].extend([gen[i]['x']['ub'] for j in range(n_s)])
    if acdc:
        for i in inc['e']:
            if 'dc' in gen[i]['output']:
                max_out['e'][i] = [z * dc_ac for z in max_out['dc'][i]]
        for i in inc['dc']:
            if 'e' in gen[i]['output']:
                max_out['dc'][i] = [z * ac_dc for z in max_out['e'][i]]
    spare_capacity = {}
    for k in s:
        for i in inc[k]:
            if k == 'h' and gen[i]['type'] == 'CombinedHeatPower':
                spare_gen[k] = [spare_gen[k][t] + (max_out[k][i][t + 1] - heat_output[i][t + 1]) for t in range(n_s)]
            elif k == 'e' and 'dc' in gen[i]['output']:
                spare_gen[k] = [spare_gen[k][t] + (max_out[k][i][t + 1] - gen_output[i][t + 1]) * dc_ac for t in range(n_s)]
            elif k == 'dc' and 'e' in gen[i]['output']:
                spare_gen[k] = [spare_gen[k][t] + (max_out[k][i][t + 1] - gen_output[i][t + 1]) * ac_dc for t in range(n_s)]
            else:
                spare_gen[k] = [spare_gen[k][t] + (max_out[k][i][t + 1] - gen_output[i][t + 1]) for t in range(n_s)]
        for i in range(n_g):
            if gen[i]['type'] == 'Utility' and k in gen[i]['output']:
                spare_capacity[k] = [spare_gen[k][t] + (gen[i]['x']['ub'] - gen_output[i][t + 1])for t in range(n_s)]
        if not k in spare_capacity:
            spare_capacity[k] = [spare_gen[k][t] + 0 for t in range(n_s)]
    return max_out, constraint, spare_capacity, spare_gen


def chp_heat(gen, output):
    n_s = len(output)
    # Assuming all outputs are nx1.
    heat_out = [
        -gen['const_demand']['district_heat'] * int(output[t] > 0)
        for t in range(n_s)
    ]
    states = gen['states'][-1]
    for t in range(n_s):
        net_out = 0
        j = 0
        # Use < instead of <= because index is 1 less than number of
        # states.
        while (j < len(states) and net_out < output[t]):
            seg = min([output[t] - net_out, gen[states[j]]['ub'][1]])
            net_out += seg
            heat_out[t] += seg * gen['output']['h'][-1][j]
            j += 1

    return heat_out



def natural_convection(normal,dT,nat_buoyancy,method):
    n = len(dT)
    if method =='DOE-2':
        # h = [max(.1,1.77*abs(dT[i])**(1/4)) for i in range(n)] # page 538 https://www.researchgate.net/publication/264308543_COMPARISON_OF_ENERGYPLUS_AND_DOE-2_DETAILED_WINDOW_HEAT_TRANSFER_MODELS
        h = [max(.1,1.31*abs(dT[i])**(1/3)) for i in range(n)] #ASHRAE vertical surface %https://www.nrel.gov/docs/fy12osti/55787.pdf page 18
    elif method =='Detailed' or method =='BLAST' or method =='TARP':
        ## TARP method %A) simple buoyancy
        h = [max(0.1,1.31*abs(dT[i])**(1/3))for i in range(n)] #ASHRAE vertical surface
        for i in range(len(dT)):
            cos_phi = normal[i][2]/(normal[i][0]**2 + normal[i][1]**2 + normal[i][2]**2)**.5 #portion of surface pointed upwards
            if (normal[i][2]<-1e-3 and dT[i]>0) or (normal[i][2]>1e-3 and dT[i]<0): #stable
                h[i] = max(0.1,1.81*abs(dT[i])**(1/3)/(1.382-abs(cos_phi))) #Walton for tilted surfaces
            elif (normal[i][2]<-1e-3 and dT[i]<0) or (normal[i][2]>1e-3 and  dT[i]>0): #unstable
                h[i] = max(0.1,9.482*abs(dT[i])**(1/3)/(7.283-abs(cos_phi))) #Walton for tilted surfaces
    else:
        ## other methods see pg 105 of reference
        print('need the adaptive convection algorithm')
    return h


def thermostat_humidistat(building,profile,T_zone):
    ## HVAC temperature Schedules
    T_set = {}
    T_set['heat'] = [T_zone[i] for i in range(len(T_zone))]
    T_set['cool'] = [T_zone[i] for i in range(len(T_zone))]
    w_set = {}
    w_set['humidify'] = [0 for i in range(len(T_zone))]
    w_set['dehumidify'] = [1 for i in range(len(T_zone))]
    T_set['no_hvac'] = [True for i in range(len(building.zones['name']))]
    for i in building.zone_controls:
        z = building.zones['name'].index(building.zone_controls[i]['zone'])
        if building.zone_controls[i]['type']=='Thermostat':
                T_set['no_hvac'][z] = False
                ct_now = profile[building.zone_controls[i]['control_type_schedule']]
                c_avail = building.zone_controls[i]['control_type']
                c_type = []
                for j  in c_avail:
                    if j=='thermostatsetpoint:singleheating':
                            c_type.append(1)
                    elif j=='thermostatsetpoint:singlecooling':
                            c_type.append(2)
                    elif j=='thermostatsetpoint:singleheatingorcooling':
                            c_type.append(3)
                    elif j=='thermostatsetpoint:dualsetpoint':
                            c_type.append(4)
                k = c_type.index(ct_now)
                c = building.zone_controls[i]['control_name'][k]
                if  c_avail[k]== 'thermostatsetpoint:singleheating':
                        T_set['heat'][z] = profile[building.thermostat[c]['heating']]
                elif c_avail[k] == 'thermostatsetpoint:singlecooling':
                        T_set['cool'][z] = profile[building.thermostat[c]['cooling']]
                elif c_avail[k] == 'thermostatsetpoint:singleheatingorcooling':
                        T_set['heat'][z] = profile[building.thermostat[c]['heating']]
                        T_set['cool'][z] = profile[building.thermostat[c]['heating']]
                elif c_avail[k] == 'thermostatsetpoint:dualsetpoint':
                        T_set['heat'][z] = profile[building.thermostat[c]['heating']]
                        T_set['cool'][z] = profile[building.thermostat[c]['cooling']]
        elif building.zone_controls[i]['type']=='Humidistat':
            w_set['humidify'][z] = profile[building.zone_controls[i]['humidify_schedule'][0]]
            w_set['dehumidify'][z] = profile[building.zone_controls[i]['dehumidify_schedule'][0]]
    T_set['dual_sp'] = False
    for i in range(len(T_zone)):
        if T_set['heat'] < T_set['cool']:
            T_set['dual_sp'] = True
    return T_set,w_set


"""Path specifications."""

import pathlib

from .user import USER_DIR


# Suffixes.
CSV_SUFFIX = '.csv'
EPW_SUFFIX = '.epw'
HDF5_SUFFIX = '.h5'
IDF_SUFFIX = '.idf'
EIO_SUFFIX = '.eio'

# File names.
DATABASE_FILENAME = 'projects'
DEFAULT_OUTPUT_FILENAME = 'output'
DISPATCH_RESULT_FILENAME = 'dispatch_result.html'

# Paths.
TLD_ABSPATH = pathlib.Path(__file__).parents[2]
EXCEL_INTERFACE_PROJECT_HELP = (
    TLD_ABSPATH / 'eagers' / 'read' / 'excel_interface_project_help.txt'
)
DEMO_DIR = TLD_ABSPATH / 'demo_files'
USER_DIR_PATH = pathlib.Path.home() / USER_DIR
USER_DIR_DATA_RETRIEVAL = USER_DIR_PATH / 'data_retrieval'
USER_DIR_EPLUS = USER_DIR_PATH / 'eplus_validation'
USER_DIR_HYDRO_DATA = USER_DIR_PATH / 'hydro_data'
USER_DIR_SIMRESULTS = USER_DIR_PATH / 'simulation_results'
USER_DIR_DATASETS = USER_DIR_PATH / 'data_sets'
USER_DIR_HDF5_DATASETS = USER_DIR_DATASETS / 'HDF5'
USER_DIR_BUILDINGS = USER_DIR_PATH / 'eplus_buildings'
USER_DIR_ENERGYPLUS_WEATHER = USER_DIR_PATH / 'eplus_weather'
USER_DIR_PROJECTS = USER_DIR_PATH / 'projects'
USER_DIR_TESTDATA = USER_DIR_PATH / 'test_data'
AUTO_CREATED_USER_DIRS = [
    USER_DIR_DATA_RETRIEVAL,
    USER_DIR_EPLUS,
    USER_DIR_HYDRO_DATA,
    USER_DIR_SIMRESULTS,
    USER_DIR_DATASETS,
    USER_DIR_HDF5_DATASETS,
    USER_DIR_BUILDINGS,
    USER_DIR_ENERGYPLUS_WEATHER,
    USER_DIR_PROJECTS,
    USER_DIR_TESTDATA,
]



"""Functionality for the Excel interface.

Class definitions:
BaseTemplateHandler
ProjectTemplateHandler
TestDataTemplateHandler
BaseTemplateWriter
ProjectTemplateWriter
TestDataTemplateWriter
BaseTemplateReader
ProjectTemplateReader
TestDataTemplateReader
"""

from abc import ABC, abstractmethod
import json

import numpy as np
from openpyxl import load_workbook, Workbook
from openpyxl.drawing.image import Image
from eagers.class_definition.building import Building
from eagers.config.dataset import TIMESTAMP_RESOLUTION_TABLENAMES
from eagers.config.excel_interface import (
    AUTO_STARTDATE_VALUES, HELP_SECTIONHEADER_WIDTH, HELP_TAB_COLOR, STYLES,
    WORKBOOK_SUFFIX)
from eagers.config.path_spec import (
    EPW_SUFFIX, HDF5_SUFFIX, EXCEL_INTERFACE_PROJECT_HELP, DEMO_DIR,
    USER_DIR_BUILDINGS, USER_DIR_ENERGYPLUS_WEATHER, USER_DIR_HDF5_DATASETS,
    USER_DIR_PROJECTS, USER_DIR_TESTDATA,
)
from eagers.config.text import RUNTIME_HDF5_FILE_PREFIX
from eagers.read.weather_file import read_weather_file
from eagers.basic.datetime_ext import datetime_from_data
from eagers.basic.file_handling import ensure_suffix
from eagers.basic.hdf5 import DatetimeFloatConverter as DFC
from eagers.basic.hdf5 import  new_hdf5_file, write_records_to_new_hdf5_table, write_user_attributes


class BaseTemplateHandler(ABC):
    """Defines functionality common between BaseTemplateWriter and
    BaseTemplateReader.
    """

    _sheet_properties = {}

    def __init__(self):
        self._filename = self._default_filename

    @property
    @classmethod
    @abstractmethod
    def _default_filename(cls):
        pass

    @property
    @classmethod
    @abstractmethod
    def _save_dir(cls):
        pass

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, value):
        if value:
            self._filename = value
        else:
            self._filename = self._default_filename

    @property
    def filepath(self):
        return ensure_suffix(
            self._save_dir / self.filename, WORKBOOK_SUFFIX)


class ProjectTemplateHandler(BaseTemplateHandler):
    _default_filename = 'project_template'
    _save_dir = USER_DIR_PROJECTS


class TestDataTemplateHandler(BaseTemplateHandler):
    _default_filename = 'testdata_template'
    _save_dir = USER_DIR_TESTDATA


class BaseTemplateWriter(BaseTemplateHandler):
    """Base template writer class."""

    _sheet_order = []

    def write(self, filename=None):
        """Write template to file.

        Keyword arguments:
        filename - (str) (Default: None) Custom file name.
        """
        # Create new workbook.
        wb = Workbook()

        # Add styles.
        for style in STYLES:
            wb.add_named_style(style)

        # Add sheets.
        for i, sheet_name in enumerate(self._sheet_order):
            self._add_sheet(wb, sheet_name, i)

        # Save template file.
        # Use property magic to update the filepath by setting the
        # filename.
        self.filename = filename
        wb.save(self.filepath)

    @classmethod
    def write_template(cls, filename=None):
        """Convenience method that requires no instantiation. Call
        signature is identical to write().
        """
        writer = cls()
        return writer.write(filename)

    @classmethod
    def _add_sheet(cls, wb, name, idx):
        """Generalized sheet creation operations.

        Positional arguments:
        wb - (Workbook) Workbook to add sheet to.
        name - (str) Name of sheet to add.
        idx - (int) Index of the new sheet.
        """
        # Generic operations.
        if idx:
            ws = wb.create_sheet(name)
        else:
            ws = wb.active
            ws.title = name
        ws.sheet_properties.tabColor = cls._sheet_properties[name].get(
            'tab_color')
        # Call method for specific operations.
        return getattr(cls, cls._sheet_properties[name]['add_method'])(ws)


class ProjectTemplateWriter(ProjectTemplateHandler, BaseTemplateWriter):
    """Project template writer class."""

    # Sheet order makes up for the lack of ordering in dictionaries
    # (changed in Python 3.7: dict class now remembers insertion order).
    _sheet_order = ['Components', 'Buildings', 'FluidLoops', 'Network',
        'TimeSeriesInputs', 'Instructions']
    _sheet_properties = {
        'Components': dict(
            add_method='_build_components_sheet',
        ),
        'Buildings': dict(
            add_method='_build_buildings_sheet',
        ),
        'FluidLoops': dict(
            add_method='_build_fluidloops_sheet',
        ),
        'Network': dict(
            add_method='_build_network_sheet',
        ),
        'TimeSeriesInputs': dict(
            add_method='_build_timeseriesinputs_sheet',
        ),
        'Instructions': dict(
            tab_color=HELP_TAB_COLOR,
            add_method='_build_help_sheet',
        ),
    }

    @classmethod
    def _build_components_sheet(cls, ws):
        """Build the given components Worksheet."""
        # Automatically procure Component classes and their headers,
        # then add them alphabetically.
        headers = cls._find_component_headers()
        for comp_name in sorted(headers):
            ws.append([comp_name])
            ws.append(headers[comp_name])
            ws.append([])
            ws.append([])
        # Formatting.
        n_header_rows = len(headers)
        period = 4
        for i_row in range(1, n_header_rows*period + 1, period):
            for row in ws[f'{i_row}:{i_row+1}']:
                for cell in row:
                    cell.style = 'header'
        ws.column_dimensions['A'].width = 20

    @classmethod
    def _build_buildings_sheet(cls, ws):
        """Build the given buildings Worksheet."""
        # Data.
        ws.append(['IDF file names'])
        # Formatting.
        ws['A1'].style = 'header'
        ws.column_dimensions['A'].width = 15

    @classmethod
    def _build_fluidloops_sheet(cls, ws):
        """Build the given fluid loops Worksheet."""
        # Data.
        headers = ['name', 'node', 'nominal_return_temperature', 'nominal_supply_temperature', 
                    'fluid_capacity', 'fluid_capacitance', 'pump_power_per_kgs', 'temperature_range']
        ws.append(headers)
        # Formatting.
        for cell in ws['1:1']:
            cell.style = 'header'
        widths = [*[9]*2, *[27]*5]
        for cell, width in zip(ws.iter_cols(max_row=1), widths):
            ws.column_dimensions[cell[0].column_letter].width = width

    @classmethod
    def _build_network_sheet(cls, ws):
        """Build the given network Worksheet."""
        headers = ['name', 'network_name', 'equipment', 'buildings', 'latitude', 'longitude', 'time_zone']
        n_connections = 10
        cnxn_headers = ['node_name','line_limit','line_efficiency']
        # Use shorter string "Connexn" to fit number in cell.
        headers.extend([f"Connexn{i}:{h}"
            for i in range(1, n_connections+1)
            for h in cnxn_headers])
        ws.append(headers)
        # Formatting.
        for cell in ws['1:1']:
            cell.style = 'header'
        widths = [*[9]*2, *[28]*3]
        for cell, width in zip(ws.iter_cols(max_row=1), widths):
            ws.column_dimensions[cell[0].column_letter].width = width

    @classmethod
    def _build_timeseriesinputs_sheet(cls, ws):
        """Build the given time series inputs Worksheet."""
        # Data.
        ws.append(['daily', 'nat_gas_price', '<- These are examples.'])
        # Formatting.
        for row in ws['A1:H1']:
            for cell in row:
                cell.style = 'header'
        widths = [13]*2
        for cell, width in zip(ws.iter_cols(max_row=1), widths):
            ws.column_dimensions[cell[0].column_letter].width = width

    @classmethod
    def _build_help_sheet(cls, ws):
        """Build the given help Worksheet."""
        # Formatting.
        ws.sheet_view.showGridLines = False
        # Data.
        ws.add_image(Image('EagersLogo.png'), 'B2')
        # Text.
        text_start_row = 10
        section_header = False
        start_text = False
        with open(EXCEL_INTERFACE_PROJECT_HELP, 'r') as helpfile:
            i_row = text_start_row
            for line in helpfile.readlines():
                line = line.rstrip('\n ')
                if not line:
                    if not start_text:
                        # Text hasn't started yet.
                        continue
                elif line[:2] == '//':
                    # Comment.
                    continue
                elif line[0] == '#':
                    # Section header.
                    line = line.strip('# ')
                    section_header = True
                start_text = True
                cell = ws[f'B{i_row}']
                cell.value = line
                if section_header:
                    for i_col in range(2, 2 + HELP_SECTIONHEADER_WIDTH):
                        ws.cell(i_row, i_col).style = 'help_sectionheader'
                    section_header = False
                i_row += 1


class TestDataTemplateWriter(TestDataTemplateHandler, BaseTemplateWriter):
    """TestData template writer class."""
    pass


class BaseTemplateReader(BaseTemplateHandler):
    """Base template reader class."""

    # Sheets that aren't read directly (or at all).
    # 'TimeSeriesInputs' is read indirectly as a result of reading other
    # sheets.
    # 'Instructions' is just for user reference.
    _ignored_sheets = ('TimeSeriesInputs', 'Instructions')

    def read(self, filename=None):
        """Read user data from filled in template.

        Keyword arguments:
        filename - (str) (Default: None) Custom file name.
        """
        # Load workbook for reading.
        # Use property magic to update the filepath by setting the
        # filename.
        self.filename = filename
        # Use "data_only" flag to avoid reading cell formulas.
        # https://stackoverflow.com/a/35624928/7232335
        if filename == 'default_project':
            wb = load_workbook(DEMO_DIR / 'default_project.xlsx', read_only=True, data_only=True)
        else:
            wb = load_workbook(self.filepath, read_only=True, data_only=True)

        # Read sheets.
        workbook_data = {}
        for sheet_name in wb.sheetnames:
            if sheet_name in self._ignored_sheets:
                continue
            workbook_data[sheet_name] = self._read_sheet(wb, sheet_name)
        return workbook_data

    @classmethod
    def read_userfile(cls, filename=None):
        """Convenience method that requires no instantiation. Call
        signature is identical to read().
        """
        reader = cls()
        return reader.read(filename)

    @classmethod
    def _read_sheet(cls, wb, name):
        """Generalized sheet reading operations.

        Positional arguments:
        wb - (Workbook) Workbook to read sheet from.
        name - (str) Name of sheet to read.
        """
        # Generic operations.
        ws = wb[name]
        # Call method for specific operations.
        read_method = cls._sheet_properties[name]['read_method']
        try:
            # Assume the read method is callable.
            return read_method(wb, ws)
        except TypeError:
            # If not callable, find it as an attribute.
            return getattr(cls, read_method)(wb, ws)

    @classmethod
    def _is_empty(cls, row):
        """Check whether the given row is empty."""
        return all(x==None for x in row)

    @classmethod
    def _rstrip_nones(cls, row):
        """Strip None values from the end (right side) of the given row.
        """
        i_startstrip = None
        for i, value in enumerate(reversed(row)):
            if value is None:
                i_startstrip = -i
            else:
                break
        if i_startstrip is None:
            # No trailing Nones values found.
            return row
        return row[:i_startstrip-1]


class ProjectTemplateReader(ProjectTemplateHandler, BaseTemplateReader):
    """Project template reader class."""

    _sheet_properties = {
        'Components': dict(
            read_method='_read_components_sheet',
        ),
        'Buildings': dict(
            read_method='_read_buildings_sheet',
        ),
        'FluidLoops': dict(
            read_method='_read_fluidloops_sheet',
        ),
        'Network': dict(
            read_method='_read_network_sheet',
        ),
        'Options': dict(
            read_method='_read_optimoptions_sheet',
        ),
    }

    def read(self, *args, **kwargs):
        """Overwrite BaseTemplateReader.read()."""
        # First do what the parent method does - read workbook data.
        workbook_data = super().read(*args, **kwargs)
        # Then bundle the project.
        proj_data = dict(
            # name: Remove file suffix, if any.
            name=self.filename.split('.')[0],
            plant=dict(
                generator=workbook_data['Components'],
                building=workbook_data['Buildings'],
                fluid_loop=workbook_data['FluidLoops'],
                network=workbook_data['Network'],
            ),
            options=workbook_data['Options'],
        )
        return proj_data

    @classmethod
    def _read_components_sheet(cls, wb, ws):
        """Read the given components Worksheet."""
        components = []
        type_row = True
        header_row = False
        for row in ws.values:
            if cls._is_empty(row):
                # Blank row.  Next non-empty row contains Component
                # type.
                type_row = True
            elif type_row:
                comp_type = row[0]
                type_row = False
                # Next row contains headers.
                header_row = True
            elif header_row:
                headers = cls._rstrip_nones(row)
                header_row = False
            else:
                comp_info = cls._interpret_components_data(wb, headers, row)
                comp_info.update(_type=comp_type)
                components.append(comp_info)
        return components

    @classmethod
    def _read_buildings_sheet(cls, wb, ws):
        """Read the given buildings Worksheet."""
        values = ws.values
        # First row is header (we know the following rows are file
        # names).
        next(values)
        buildings = []
        for row in values:
            filename = row[0]
            try:
                build = Building.from_file(DEMO_DIR / filename)
            except:
                build = Building.from_file(USER_DIR_BUILDINGS / filename)
            build.filename = filename
            buildings.append(build)
        return buildings

    @classmethod
    def _read_fluidloops_sheet(cls, wb, ws):
        """Read the given fluidloops Worksheet."""
        values = ws.values
        headers = cls._rstrip_nones(next(values))
        fluidloops = []
        for row in values:
            row = cls._rstrip_nones(row)
            if row:
                fluidloops.append({
                    key: val
                    for key, val in zip(headers, cls._rstrip_nones(row))
                })
        return fluidloops

    @classmethod
    def _read_network_sheet(cls, wb, ws):
        """Read the given network Worksheet."""
        headers = [cell.value for cell in ws[1]]
        networks = []
        for row in ws.iter_rows(min_row=2, values_only=True):
            row = cls._rstrip_nones(row)
            if not row:
                # The row has no data in it.
                continue
            net = {
                key: val
                for key, val in zip(headers, row)
                if val is not None
            }
            if 'equipment' in net:
                net['equipment'] = net['equipment'].split(",")
                net['equipment'] = [i.strip() for i in net['equipment']]
            if 'buildings' in net:
                net['buildings'] = net['buildings'].split(",")
                net['buildings'] = [i.strip() for i in net['buildings']]
            networks.append(net)
        return networks

    @classmethod
    def _read_optimoptions_sheet(cls, wb, ws):
        """Read the given optimoptions worksheet."""
        optimoptions = dict(cls._rstrip_nones(x)
            for x in ws.iter_rows(min_row=2, values_only=True))
        # Special handling of start_date value.
        if optimoptions['start_date'] in AUTO_STARTDATE_VALUES:
            optimoptions['start_date'] = None
        else:
            optimoptions['start_date'] = \
                datetime_from_data(optimoptions['start_date'])
        return optimoptions

    @classmethod
    def _interpret_components_data(cls, wb, headers, data):
        """Return a dictionary of data built from the given headers and
        row of data. This will convert strings to the actual values they
        represent.

        Positional arguments:
        wb - (Workbook) Applicable Workbook.
        headers - (iterable of strings) Header names.
        data - (iterable of data values) Data read from a row of a
            components worksheet.
        """
        # Interpreting function.
        def interpret_value(v):
            # Multiline values: Remove newlines.
            if isinstance(v, str):
                v = v.replace('\n', '')
            # Try using the JSON module to decode values.  This is
            # mostly useful for lists.
            # Can encounter TypeError (if value is not a string) or
            # JSONDecodeError (if value is a string but isn't
            # understood).
            try:
                return json.loads(v)
            except TypeError:
                return v
            except json.JSONDecodeError:
                # Accept "inf" and "-inf" as +/- Infinity.
                if v in ('inf', '-inf'):
                    return float(v)
                return v

        # Build data structure.
        multilevel_field = None
        comp_info = {}
        for h, v in zip(headers, data):
            if '.' in h:
                # NOTE: Assumes only one '.' present, so only 2 levels
                # of compounding are supported.
                field, h = h.rsplit('.', maxsplit=1)
                if multilevel_field != field:
                    comp_info[field] = {}
                info_structure = comp_info[field]
                multilevel_field = field
            else:
                info_structure = comp_info
            if ':' in h and v is not None:
                sheetname, h = h.split(':')
                v_split = v.split(',')
                ts_header, data_header = v_split[0].strip(), v_split[1].strip()
                time, values = cls._read_timeseries_input(
                    wb[sheetname], ts_header, data_header)
                info = {'timestamp': time, h: values}
            else:
                info = {h: interpret_value(v)}
            if None not in info.values():
                info_structure.update(info)
        return comp_info

    @classmethod
    def _read_timeseries_input(cls, ws, ts_header, data_header):
        """Read a field from a time series worksheet.

        Positional arguments:
        ws - (Worksheet) Applicable Worksheet.
        ts_header - (str) Name of time series header.
        data_header - (str) Name of data header.
        """
        # Get worksheet values generator.
        values = ws.values
        # Read first row of values as headers.
        headers = cls._rstrip_nones(next(values))
        i_ts = headers.index(ts_header)
        i_data = headers.index(data_header)
        # Read remaining rows as data.
        time = []
        data = []
        for row in values:
            time.append(row[i_ts])
            data.append(row[i_data])
        return time, data


class TestDataTemplateReader(TestDataTemplateHandler, BaseTemplateReader):
    """TestData template reader class."""

    _sheet_properties = {
        'NodeData': dict(
            read_method='_read_nodedata_sheet',
        ),
        'Weather': dict(
            read_method='_read_weather_sheet',
        ),
    }

    def read(self, filename=None):
        """Overwrite BaseTemplateReader.read(). Read user data from
        filled in TestData file template.

        Keyword arguments:
        filename - (str) (Default: None) Custom file name.
        """
        # Load workbook for reading.
        # Use property magic to update the filepath by setting the
        # filename.
        self.filename = filename
        # Use "data_only" flag to avoid reading cell formulas.
        # https://stackoverflow.com/a/35624928/7232335
        if filename == 'default_testdata':
            wb = load_workbook(DEMO_DIR / 'default_testdata.xlsx', read_only=True, data_only=True)
        else:
            wb = load_workbook(self.filepath, read_only=True, data_only=True)

        # Discover sheets.
        datacat_sheets = [sh for sh in wb.sheetnames
            if sh not in self._ignored_sheets]
        data_categories = [sh.lower() for sh in datacat_sheets]

        # Read sheets for TimeSeriesInputs headers and other info.
        sheet_data = {cat: self._read_sheet(wb, sh)
            for cat, sh in zip(data_categories, datacat_sheets)}

        # Read weather file, if necessary.
        weatherfile_name = sheet_data['weather']['filename']
        weatherfile_designation = sheet_data['weather']['designation']
        weather_data = self._read_weatherfile_data(
            weatherfile_name, weatherfile_designation)
        if weatherfile_name:
            # Delete the weather entry so it isn't used in the upcoming
            # reading of TimeSeriesInputs.
            del sheet_data['weather']

        # Read header information from TimeSeriesInputs sheet.
        table_info, from_tsi = self._read_timeseriesinputs_header_info(
            wb['TimeSeriesInputs'], sheet_data)

        # Add weather info from file, if relevant.
        if weatherfile_name:
            tstamp = weather_data['timestamp']
            ts_resolution = tstamp[1] - tstamp[0]
            designation = weather_data.pop('designation')
            from_tsi['weather'] = False
            table_info['weather'] = {
                TIMESTAMP_RESOLUTION_TABLENAMES[ts_resolution]: dict(
                    designation=designation,
                    data=weather_data,
                ),
            }

        # Generate table descriptions.
        table_descriptions = \
            self._generate_table_descriptions(table_info, from_tsi)

        # Generate file names.
        file_info = dict(
            {cat: dict(
                file=(USER_DIR_HDF5_DATASETS / cat
                    / f"{RUNTIME_HDF5_FILE_PREFIX}{cat}{HDF5_SUFFIX}"),
            ) for cat in data_categories}
        )

        # Write data to HDF5.
        self._write_tsi_data_to_hdf5(
            wb['TimeSeriesInputs'],
            table_info,
            file_info,
            table_descriptions,
            from_tsi,
        )

        # Gather all return information.
        for cat in file_info:
            # Remove preamble to file paths.
            file_info[cat]['filename'] = file_info[cat].pop('file').name
            try:
                sheet_data[cat].update(file_info[cat])
            except KeyError:
                sheet_data[cat] = file_info[cat]
        # Flatten.
        return {f'{cat}_{field}': data2
            for cat, data1 in sheet_data.items()
                for field, data2 in data1.items()}

    @classmethod
    def _read_nodedata_sheet(cls, wb, ws):
        """Read the given node data Worksheet."""
        # Get worksheet values generator.
        values = ws.values
        # First row is headers.
        headers = next(values)
        hydro_only_headers = \
            {'inflow', 'outflow', 'inflow_history', 'outflow_history'}
        # Read remaining rows as data.
        tsi_headers = {}
        network_info = {}
        for row in values:
            if cls._is_empty(row):
                # Blank row.  Do nothing.
                continue
            else:
                row = cls._rstrip_nones(row)
                if not row[0] is None:
                    net = row[0]
                else:
                    row = list(row)
                    row[0] = net
                if len(row) > 1:
                    # More than just a network name.
                    row_data = dict(zip(headers, row))
                    # Check hydro-specific data requirements.
                    # hydro_check = row_data['network'] == 'hydro' \
                    #     and set(row_data.keys()) == set(headers)
                    nonhydro_check = net != 'hydro' \
                        and not (set(row_data.keys())
                            .intersection(hydro_only_headers))
                    assert nonhydro_check, \
                        ("Hydro-specific data must be found only in only hydro"
                        "network rows.")
                    # Append to set of TimeSeriesInputs headers.
                    for k, v in row_data.items():
                        if ',' not in v:
                            continue
                        tstamp, field = v.replace(' ', '').split(',')
                        try:
                            tsi_headers[tstamp].add(field)
                        except KeyError:
                            tsi_headers[tstamp] = {field}
                        row_data[k] = (tstamp, field)
                    # Append to list of node data.
                    try:
                        network_info[row_data['network']].append(row_data)
                    except KeyError:
                        network_info[row_data['network']] = [row_data]
        return dict(
            tsi_headers=tsi_headers,
            network_info=network_info,
        )

    @classmethod
    def _read_weather_sheet(cls, wb, ws):
        """Read the given weather Worksheet."""
        # Get worksheet values generator.
        values = ws.values
        # Look for file name and designation in first two rows.
        filename = next(values)[1]
        designation = next(values)[1]
        if filename:
            return dict(
                tsi_headers=None, designation=designation, filename=filename)
        # Read remaining rows as data.
        tsi_headers = {}
        tsi_row = 1
        ts_field = None
        for row in values:
            if cls._is_empty(row):
                # Empty row.  Next non-empty row should be timestamp.
                tsi_row = 1
                continue
            elif tsi_row == 1:
                ts_field = row[1]
                tsi_row += 1
            elif ts_field is None:
                # A timestamp field was not specified.
                tsi_row = 0
            elif tsi_row == 2:
                tsi_headers[ts_field] = set(cls._rstrip_nones(row)[1:])
                tsi_row += 1
            elif tsi_row == 3:
                designation = row[1]
                tsi_row += 1
        return dict(
            tsi_headers=tsi_headers,
            designation=designation,
            filename=None,
        )

    @classmethod
    def _read_weatherfile_data(cls, filename, designation):
        if filename:
            try:
                weather_data = read_weather_file(ensure_suffix(DEMO_DIR / filename, EPW_SUFFIX,), designation,)
            except:
                weather_data = read_weather_file(ensure_suffix(USER_DIR_ENERGYPLUS_WEATHER / filename, EPW_SUFFIX,), designation,)
            return weather_data

    @classmethod
    def _read_timeseriesinputs_header_info(cls, ws, sheet_data):
        tsi_headers = cls._rstrip_nones([c.value for c in ws['1']])
        # Indices must be sorted so that read order matches column order
        # in the HDF5 file.
        # TimeSeriesInputs headers are prefixed with "<data category>:"
        # so that it's possible to have the same header name in two
        # different categories.
        tsi_headers_set = set(tsi_headers)
        from_tsi = {}
        table_info = {}
        for cat, data in sheet_data.items():
            from_tsi[cat] = True
            table_info[cat] = {}
            for tablename, hdrs in data.pop('tsi_headers').items():
                h_cat = [tablename, *(f"{cat}:{x}" for x in hdrs)]
                assert set(h_cat) <= tsi_headers_set, \
                    ("Not all named headers appear on the TimeSeriesInputs "
                    "sheet.")
                # Sort headers based on the order in which they appear
                # on the TimeSeriesInputs sheet.
                i_sorted, h_sorted = zip(*sorted(
                    ((i, h) for i, h in enumerate(tsi_headers) if h in h_cat),
                    key=lambda x: x[0],
                ))
                table_info[cat][tablename] = dict(
                    indices=i_sorted,
                    names=h_sorted,
                )
                if cat == 'weather':
                    table_info[cat][tablename]['designation'] = \
                        sheet_data['weather'].pop('designation')
        return table_info, from_tsi

    @classmethod
    def _generate_table_descriptions(cls, table_info, from_tsi):
        table_descriptions = {}
        for cat, data in table_info.items():
            table_descriptions[cat] = {}
            if from_tsi[cat]:
                for table, h_info in data.items():
                    table_descriptions[cat][table] = {}
                    dtype = []
                    for h in h_info['names']:
                        h_split = h.split(':')
                        try:
                            h = h_split[1]
                        except IndexError:
                            h = 'timestamp'
                        dtype.append((h, 'f8'))
                    table_descriptions[cat][table] = np.dtype(dtype)
            else:
                for table, h_info in data.items():
                    table_descriptions[cat][table] = {}
                    dtype = []
                    for h in h_info['data']:
                        dtype.append((h, 'f8'))
                    table_descriptions[cat][table] = np.dtype(dtype)
        return table_descriptions

    @classmethod
    def _gather_data_rows(cls, ws, i_headers, dtype):
        """Returns a structured array of data values for a given set of
        headers.
        """
        n_pts = ws.max_row - 1
        data = np.empty(dtype=dtype, shape=n_pts)
        timestamp = np.empty(dtype='O', shape=n_pts)
        i_headers = np.array(i_headers)
        non_ts = np.array(dtype.names) != 'timestamp'
        i_non_ts = i_headers[non_ts]
        i_ts = i_headers[np.invert(non_ts)][0]
        # ReadOnlyWorksheet does not have iter_cols().  Use
        # iter_rows().
        for i, row in enumerate(ws.iter_rows(min_row=2, values_only=True)):
            row = cls._rstrip_nones(row)
            data[i] = (0, *(row[j] for j in i_non_ts))
            timestamp[i] = datetime_from_data(row[i_ts])
        # Convert datetimes to floats; this is the only way PyTables
        # can store datetime values.
        data['timestamp'] = DFC.d2f_arr2arr(timestamp)
        return data

    @classmethod
    def _write_tsi_data_to_hdf5(cls,
            ws, table_info, file_info, table_descriptions, from_tsi):
        for cat, data in table_info.items():
            new_hdf5_file(file_info[cat]['file'], overwrite=True)
            for table, h_info in data.items():
                if from_tsi[cat]:
                    records = cls._gather_data_rows(
                        ws,
                        h_info['indices'],
                        table_descriptions[cat][table],
                    )
                else:
                    # Use a key, not a value; that would copy all the
                    # data into a list.
                    rando_key = list(h_info['data'].keys())[0]
                    records = np.empty(
                        dtype=table_descriptions[cat][table],
                        shape=len(h_info['data'][rando_key]),
                    )
                    for k, v in h_info['data'].items():
                        if k == 'timestamp':
                            v = DFC.d2f_lst2arr(v)
                        records[k] = v
                write_records_to_new_hdf5_table(
                    file_info[cat]['file'],
                    table,
                    records,
                )
                if cat == 'weather':
                    # Write designation to file.
                    write_user_attributes(
                        file_info[cat]['file'],
                        table,
                        dict(designation=h_info['designation']),
                    )



"""Configuration variables for various types of text.
"""

# FILE NAME INDICATORS
VERSION_INDICATOR = "v"


# NAMES
# General
# Name separator.
NAME_SEP = "_"
# Prefix for runtime HDF5 files.
RUNTIME_HDF5_FILE_PREFIX = "runtime_"

# CSV result file
# Separator between project name and table name.
CSV_RESULT_SEP = "--"

# Acceptable True/False strings.  Use str.lower() to compare.
ACCEPTABLE_TRUE_STR = ("true", "1", "y")
ACCEPTABLE_FALSE_STR = ("false", "0", "n")
def parse_bool_str(s):
    lower = s.lower()
    if lower in ACCEPTABLE_TRUE_STR:
        return True
    elif lower in ACCEPTABLE_FALSE_STR:
        return False
    else:
        raise ValueError(f"Unrecognized boolean value {s!r}")



from datetime import timedelta
from numpy import log
from scipy.stats import t

from eagers.basic.get_data import get_data
from eagers.forecasting.arima import arima
from eagers.forecasting.arma import arma
from eagers.forecasting.surface_forecast import surface_forecast
from eagers.forecasting.weather_forecast import weather_forecast


def update_forecast(options, date, test_data,subnet):
    """Update forecast using any of the following forecasting
    algorithms:
    Perfect     Perfect forecast. Requires future data, and therefore is
                only available in simulation mode.
    Surface     Surface forecast.
    ARMA        AutoRegressive Moving Average forecast.
    ARIMA       AutoRegressive Integrated Moving Average forecast.
    ANN         Artificial Neural Network forecast.
    
    Positional arguments:
    options - (Optimoptions) Project options.
    date - (list of timestamp) Timestamps for which a forecast is
        requested.
    test_data 
    subnet - (dict) Network data.
    """
    hist_prof = test_data['hist_prof']
    # Period of repetition (1 = 1 day). This is how far back the
    # forecasting methods are able to see. It is irrelevant if the
    # forecast is perfect.
    
    if not options['forecast'] in ['perfect','uncertain']:
        #TODO replace with observer history
        d0 = date[0] - timedelta(hours = 24+options['resolution'])
        prev_date = [d0 + timedelta(hours = i*options['resolution']) for i in range(round(24/options['resolution'],0)+1)]
        prev_data = get_data(test_data,prev_date,subnet['network_names'])
    forecast = {}
    forecast['timestamp'] = date
    if options['forecast'] != 'arima' and not options['forecast'] in ['perfect','uncertain']:
        forecast['weather'] = weather_forecast(test_data,prev_data, hist_prof, date)
    if options['forecast'] == 'arma':
        forecast['demand'] = arma(date, prev_data)
    elif options['forecast'] == 'arima':
        forecast = arima(date, prev_data, options)
    elif options['forecast'] == 'neural_net':
        pass
    #TODO # create neural network forecasting option
    elif options['forecast'] == 'surface':
        forecast['demand'] = surface_forecast(prev_data, hist_prof['demand'], date, forecast['weather']['t_dryb'],[])
    elif options['forecast'] in ['perfect','uncertain']:
        forecast = get_data(test_data, date, subnet['network_names'])
        if options['forecast'] == 'uncertain':
            forecast = make_uncertain(forecast,options['forecast_uncertainty'])
    elif options['forecast'] == 'building':
        pass
    else:
        raise RuntimeError('Forecast option not recognized.')
    ### Make first hour forecast "perfect"
    # make_perfect(forecast,options,test_data,date,subnet)
    if options['spin_reserve']:
        if 'demand' in forecast:
            #TODO update this
            forecast['sr_target'] += options['spin_reserve_perc'] / 100 * sum(
                forecast['demand']['e'], 2)
    return forecast

def make_perfect(forecast,options,test_data,date,subnet):
    #Make first step of forecast perfect:
    if options['method'].lower() == 'dispatch' and len(date) > 1:
        next_data = get_data(test_data, date[0], subnet['network_names'])
        for k in next_data:
            if isinstance(next_data[k], dict):
                for s_i in next_data[k]:
                    forecast[k][s_i][0] = next_data[k][s_i][0]
            else:
                forecast[k][0] = next_data[k][0]

def make_uncertain(forecast,sp):
    '''Apply forecast error to the given forecast, using the given
    random number generator.

    Positional arguments:
    forecast - (dict) Forecast information with the same structure as
        returned by get_data().
    sp - forecast scale paramater
    '''
    gamma = [float(log(i+2) * sp / 6.3138) for i in range(len(forecast['timestamp']))]
    node_names = list(forecast.keys())
    for node in node_names:
        if not node in ['weather','timestamp'] and 'demand' in forecast[node]:
            n = len(forecast[node]['demand'])
            ran = t.rvs(1, size=n)
            rscale = [max(0.5,(1+min(0.5,float(ran[i])*gamma[i]/100)))for i in range(n)]
            forecast[node]['demand'] = [forecast[node]['demand'][i]*rscale[i] for i in range(n)]
    # Weather.
    for k in ['glo_horz_irr','dir_norm_irr','dif_horz_irr','t_dryb','wspd']:
        if k in forecast['weather']:
            n = len(forecast['weather'][k])
            ran = t.rvs(1, size=n)
            rscale = [max(0.5,(1+min(0.5,float(ran[i])*gamma[i]/100)))for i in range(n)]
            forecast['weather'][k] = [forecast['weather'][k][i]*rscale[i] for i in range(n)]
    return forecast



"""Network loading logic.
"""

from eagers.config.network import NETWORK_NAMES, NETWORK_ABBRS


def load_network(network, gen, buildings):
    """Identify generators and lines and their position in the network.
    Organize into sub-networks.
    Group nodes between which transmission losses don't occur.
    Only create lines where transmission losses do occur.
    """

    subnet = {}
    n_nodes = len(network)
    node_names = []
    location = []
    network_names = []
    for i in range(n_nodes):
        node_names.append(network[i]['name'])
        if not network[i]['network_name'] in network_names:
            network_names.append(network[i]['network_name'])
        if 'location' in network[i]:
            location.append(network[i]['location'])
        else:
            location.append([])
        
    subnet['network_names'] = network_names
    for j in network_names:
        subnet[j] = {}

    line_number =  0  
    for net in subnet['network_names']:
        subnet[net]['abbreviation'] = NETWORK_ABBRS[NETWORK_NAMES.index(net)]
        subnet[net]['line'] = {}
        subnet[net]['line']['node1'] = []
        subnet[net]['line']['node2'] = []
        subnet[net]['line']['network'] = []
        subnet[net]['line']['number'] = []
        subnet[net]['line']['limit'] = []
        subnet[net]['line']['eff'] = []
        subnet[net]['line']['dir'] = []
        subnet[net]['nodes'] = []
        subnet[net]['location'] = []
        subnet[net]['load'] = []
        subnet[net]['connections'] = []
        for i in range(n_nodes):
            if network[i]['network_name'] == net:
                # First check to see if this node is already a part of
                # subnet node.
                # Nodes with perfect transmission are aggregated into
                # the first node in the name_list that they have a
                # perfect two-way connection with.
                ind = None
                for k in range(len(subnet[net]['nodes'])):
                    if node_names[i] in subnet[net]['nodes'][k]:
                        ind = k
                if ind is None:
                    ind, a_nodes, connect = aggregated_node(network, node_names[i], net, node_names)
                if ind == i:
                    subnet[net]['nodes'].append(a_nodes)  # Add a new subnet node.
                    subnet[net]['location'].append(location[i])
                    subnet[net]['load'].append([])
                    if connect is None:
                        connections = None
                    else:
                        connections = []
                        for j in range(len(connect[1])):
                            if not connect[1][j] in a_nodes:#imperfect transmission, need a line
                                J, c_nodes,_ = aggregated_node(network, connect[1][j], net,node_names)
                                p_connected = node_names[J] #name of node that the connected node will be agregated into if it is perfectly connected to any others
                                connections.append(p_connected)
                                if J>i: #new line connection, otherwise this was handled previously in the reverse direction
                                    eff, limit, direction = line_prop(network,a_nodes, c_nodes, net,node_names)
                                    if direction == 'none': #no transmission (zero efficiency)
                                        pass #do nothing
                                    else:
                                        subnet[net]['line']['dir'].append(direction)
                                        subnet[net]['line']['network'].append(net)
                                        if direction == 'reverse':
                                            subnet[net]['line']['node1'].append(p_connected)
                                            subnet[net]['line']['node2'].append(node_names[i])
                                        else:
                                            subnet[net]['line']['node1'].append(node_names[i])
                                            subnet[net]['line']['node2'].append(p_connected)
                                        subnet[net]['line']['eff'].append(eff)
                                        subnet[net]['line']['limit'].append(limit)
                                        subnet[net]['line']['number'].append(line_number)
                                        line_number += 1
                    subnet[net]['connections'].append(connections)
    subnet = create_obj_lists(network, subnet, node_names, 'equipment')
    if 'hydro' in network_names:
        subnet = add_river_segments(network, gen, subnet, line_number, location, node_names)
    if buildings:
        f_names = [b.filename for b in buildings]
        for n in network:
            if not n['buildings'] is None:
                n['buildings'] = [f_names.index(k) for k in n['buildings']]
        subnet = create_obj_lists(network,subnet,node_names,'buildings')
    return subnet


def create_obj_lists(network,subnet,node_names,o_type):
    """Identify objects at each sub-net node (objects can appear in
    multiple sub-nets if they produce heat and power or use water to
    produce electricity, but can only appear at one node in a sub-net.
    """
    for net in subnet['network_names']:
        subnet[net][o_type] = []
        for n in range(len(subnet[net]['nodes'])):
            obj_at_node = []
            node_m = subnet[net]['nodes'][n]
            for k in range(len(node_m)):
                i_net = node_names.index(node_m[k])
                if o_type == 'equipment':
                    if network[i_net]['equipment'] is not None:
                        obj_at_node.extend(network[i_net]['equipment'])
                elif o_type == 'buildings':
                    if network[i_net]['buildings'] is not None:
                        obj_at_node.extend(network[i_net]['buildings'])
            subnet[net][o_type].append(obj_at_node)
    return subnet


def line_prop(network, node_1, node_2, net,node_names):
    """Find the transmission efficiency and limit between 2 connected
    nodes. If one of the nodes is perfectly connected to another node
    there may be more than one pathway connecting them, so aggregate the
    lines.
    """
    def eff_and_limit(eff,limit,new_eff,new_limit,ind):
        if new_limit == float('inf') or limit[ind] == 0:
            eff[ind] = new_eff
            limit[ind] = new_limit
        else:
            weight = new_limit/(limit[ind] + new_limit)
            eff[ind] = (1-weight)*eff[ind] + weight*new_eff
            limit[ind] = limit[ind] + new_limit
        return eff, limit

    eff = [0 for i in range(2)]
    limit = [0 for i in range(2)]
    for j in node_1:
        name_index = node_names.index(j)
        for k in node_2:
            J = node_names.index(k)
            if k in network[name_index]['connections']['node_name'] and j in network[J]['connections']['node_name']:
                #forward direction efficiency and limit
                c = network[name_index]['connections']['node_name'].index(k)
                eff, limit = eff_and_limit(eff,limit,network[name_index]['connections']['line_efficiency'][c],network[name_index]['connections']['line_limit'][c],0)
                #reverse direction efficiency and limit
                c = network[J]['connections']['node_name'].index(j)
                eff, limit = eff_and_limit(eff,limit,network[J]['connections']['line_efficiency'][c],network[J]['connections']['line_limit'][c],1)
    if eff[0]==0 and eff[1]>0:
        direction = 'reverse'
        eff = [eff[1]]
        limit = [limit[1]]
    elif eff[0]>0 and eff[1]==0:
        direction = 'forward'
        eff = [eff[0]]
        limit = [limit[0]]
    elif eff[0]==0 and eff[1]==0:
        direction = 'none'
        eff = []
        limit = []
    else:
        direction = 'dual'
    return eff, limit, direction


def aggregated_node(network, node, net, node_names):
    # Any connected nodes with perfect bi-directional transfer are
    # aggregated into the node earliest in the list of names.
    # This function finds which node in the list that is
    name_index = node_names.index(node)
    a_nodes = [node]
    connect = None
    if network[name_index]['connections']['node_name'] != None:
        nc = len(network[name_index]['connections']['node_name'])
        connect = [[node for i in range(nc)],network[name_index]['connections']['node_name']]
        k = 0
        while k<len(connect[1]):
            if not connect[1][k] in a_nodes: #avoid looking at connections to nodes already in agregated node
                eff,_,_  = line_prop(network,a_nodes, [connect[1][k]],net,node_names)
                if len(eff)==2 and min(eff)==1 and not net == 'hydro': #perfect bi-directional energy transfer, hydro lines are river segments, can't agregate
                    J = node_names.index(connect[1][k])
                    a_nodes.append(connect[1][k])
                    new_connect = [nn for nn in network[J]['connections']['node_name'] if nn not in a_nodes]
                    connect[0].extend([connect[1][k] for i in range(len(new_connect))])
                    connect[1].extend(new_connect)
                    del connect[0][k]
                    del connect[1][k]
                    k -= 1
                    if J<name_index:
                        name_index =J#keep lowest number (index in list of node names)
            k += 1
        if len(connect[1])==0:
            connect = None
    return name_index, a_nodes, connect


def add_river_segments(network,gen,subnet,line_number,location,node_names):
    subnet['hydro'] = {}
    subnet['hydro']['nodes']  = []
    subnet['hydro']['abbreviation'] = 'w'
    subnet['hydro']['location'] = []
    subnet['hydro']['connections'] = []
    subnet['hydro']['load'] = []
    subnet['hydro']['line']['number'] = []
    subnet['hydro']['line']['minimum'] = []
    subnet['hydro']['line']['limit'] = []
    subnet['hydro']['equipment'] = []
    subnet['hydro']['line']['node1'] = []
    subnet['hydro']['line']['node2'] = []
    subnet['hydro']['line']['network'] = []
    subnet['hydro']['line']['time'] = []

    num_gen = len(gen)
    gen_names = []
    for i in range(num_gen):
        gen_names.append(gen[i]['name'])

    hydro_node = []
    down_river = []
    for i in range(len(network)):
        if network[i]['network_name'] == 'hydro':#add a new subnet node
            line_number = line_number+1
            gen_at_node = []
            equip = network[i]['equipment']
            for j in equip:
                gen_i = gen_names.index(j)
                if gen[gen_i].type()== 'HydroStorage':
                    gen_at_node.append(gen_i)
            subnet['hydro']['nodes'].append(network[i]['name'])
            subnet['hydro']['location'].append(location[i])
            subnet['hydro']['connections'].append(network[i]['hydro']['connections'])
            subnet['hydro']['load'].append([]) #not empty if there is a water diversion for irrigation etc
            subnet['hydro']['line']['number'].append(line_number)
            subnet['hydro']['line']['minimum'].append(network[i]['hydro']['instream_flow'])
            subnet['hydro']['line']['limit'].append(float('inf'))
            subnet['hydro']['equipment'].append(gen_at_node)
            subnet['hydro']['line']['network'].append('hydro')
            hydro_node.append(network[i]['name']) #node names of the upriver node (origin of line segment)
            connect_nodes = network[i]['connections']['node_name']
            if not connect_nodes is None:
                dr_node = node_names.index(connect_nodes[0])
                down_river.append(network[dr_node]['name']) #node names of the downriver node (end of line segment)
                subnet['hydro']['line']['node1'].append(network[i]['name'])
                subnet['hydro']['line']['node2'].append(network[dr_node]['name'])
                subnet['hydro']['line']['time'].append(network[i]['hydro']['time_to_sea'] - network[dr_node]['hydro']['time_to_sea']) #transit time from current river to downstream river
            else:#last dam before the sea
                subnet['hydro']['line']['node1'].append(network[i]['name'])
                subnet['hydro']['line']['node2'].append(None)
                subnet['hydro']['line']['time'].append(network[i]['hydro']['time_to_sea']) #no downstream dam
    #following indexing helps locate upstream nodes during load_matrices, and forecast_hydro
    subnet['hydro']['up_river'] = []
    for i in range(len(subnet['hydro']['nodes'])):
        subnet['hydro']['up_river'].append(down_river.index(hydro_node[i])) #finds which nodes the current node is downstream of
    return subnet


"""Functionality for finding out which node a generator is at.
"""


def find_gen_node(gen, subnet):
    """Find out which node the given generator exists at."""
    gen_in_network = {}
    for k, v in subnet.items():
        if k == "network_names":
            continue
        for i, node in enumerate(v["equipment"]):
            if gen["name"] in node:
                gen_in_network[k] = i
                break
        else:
            # break was not reached.
            gen_in_network[k] = None
    gen_network = None
    i_node = None
    for network_type in subnet["network_names"]:
        if gen_in_network[network_type] is not None:
            gen_network = network_type
            i_node = gen_in_network[network_type]
            break
    if not gen_network:
        raise RuntimeError(f"No network found for generator {gen['name']!r}")
    return gen_network, i_node



import warnings

import numpy as np
from scipy import sparse
from scipy.linalg import sqrtm
import ecos


def ecos_ls(H, f, A, b, Aeq, beq):
    ''' 
    Need to convert to second order conic problem following template for matlab ecos:
    https://github.com/embotech/ecos-matlab/blob/master/bin/ecosqp.m
    See Linear Cone Program example on https://cvxopt.readthedocs.io/en/latest/coneprog.html for creation of G and h
    Other info:
    https://math.stackexchange.com/questions/1932532/converting-from-quadratic-to-second-order-cone-optimization-problem
    https://inst.eecs.berkeley.edu/~ee227a/fa10/login/l_conic_socp.html 
    
    so the states x are expanded with y states representing xHx anytime H[i]>0 so the translation involves:
    x = [x;y] 
    c = [zeros(x); 1]
    Aeq = [Aeq,zeros(x)]
    beq = beq
    U = sqrt(H)  ** easy if H is positive semi-definite
    G = [[A , zeros(x)]             h = [[b;]
        [f'/sqrt(2), -1/sqrt(2)]        [1/sqrt(2);]
        [-U; zeros(x)]                  [zeros(x);]
        [f'/sqrt(2); -1/sqrt(2)]        [1/sqrt(2);]] (second order)
    '''
    x_l = len(f)
    #create triplets for A and A_eq
    val_Aeq = []
    row_Aeq = []
    col_Aeq = []
    req=0
    for row in range(len(Aeq)):
        for i in range(len(Aeq[row])):
            if Aeq[row][i]!=0:
                val_Aeq.append(Aeq[row][i])
                row_Aeq.append(req)
                col_Aeq.append(i)
        req += 1
    

    val_G = []
    row_G = []
    col_G = []
    r=0
    for row in range(len(A)):
        for i in range(len(A[row])):
            if A[row][i]!=0:
                val_G.append(A[row][i])
                row_G.append(r)
                col_G.append(i)
        r+= 1

    # add the first cone constraint 
    for i in range(x_l):
        val_G.append(f[i]/np.sqrt(2))
        row_G.append(r)
        col_G.append(i)
    
    val_G.append(-1/np.sqrt(2))
    row_G.append(r)
    col_G.append(x_l)

    cones = 0
    H_ar = np.array(H)
    U_ar = sqrtm(H_ar)
    with warnings.catch_warnings():
        # Filter out ComplexWarnings.  There might be some miniscule
        # imaginary part from numerical error while taking the square
        # root of a matrix.
        warnings.filterwarnings("ignore", category=np.ComplexWarning)
        U = [[float(U_ar[i,j]) for i in range(x_l)] for j in range(x_l)]
    for row in range(len(U)):
        for i in range(len(U[row])):
            if U[row][i]!=0:         
                #add the state constraint (2nd order part on LHS of inequality)
                val_G.append(-U[row][i])
                row_G.append(r+cones+1)
                col_G.append(i)
        cones += 1

    # add the other cone constraint
    for i in range(x_l):
        val_G.append(-f[i]/np.sqrt(2))
        row_G.append(r + cones + 1)
        col_G.append(i)

    val_G.append(1/np.sqrt(2))
    row_G.append(r + cones + 1)
    col_G.append(x_l)

    G = sparse.csc_matrix((np.array(val_G),(np.array(row_G),np.array(col_G))), shape = (r+cones+2,x_l+1),dtype='float64')

    h = [b[j] for j in range(len(b))]
    h.append(1/np.sqrt(2))
    h.extend([0 for j in range(cones)])
    h.append(1/np.sqrt(2))
    h = np.array(h)

    dims = {}
    dims['l'] = r 
    dims['q'] = [cones+2]
    dims['e'] = 0

    c = [float(0) for j in range(x_l)]
    c.append(1)
    c = np.array(c)

    A_ecos = sparse.csc_matrix((np.array(val_Aeq),(np.array(row_Aeq),np.array(col_Aeq))), shape = (req,x_l+1),dtype='float64')
    b_ecos = np.array([beq[i] for i in range(len(beq))])
    
    solution = ecos.solve(c, G, h, dims, A_ecos, b_ecos, verbose=False)

    # Exit Flags:
    # 0: optimal
    # 1: primal infeasible
    # 2: dual infeasible
    # 10: close to optimal solution found
    # -1: maximum number of iterations reached
    # -2: numerical problems (unreliable search direction)
    # -3: numerical problems (slacks or multipliers became exterior)
    # -7: unknown problem in solver

    if solution['info']['exitFlag']==0 or solution['info']['exitFlag'] == 10:
        feasible = 0
    else:
        feasible = solution['info']['exitFlag']
    #add removed states back in as zero
    x = [0 for i in range(x_l)]
    for i in range(x_l):
        x[i] = float(solution['x'][i])
    return x, feasible




from eagers.building.basic.psychometric import psychometric
from eagers.building.basic.eval_curve import eval_curve
from eagers.building.components.water_equipment import water_equipment
from eagers.building.basic.window_solar_transmittance import window_solar_transmittance
from eagers.building.basic.solar_calc import solar_calc

import numpy as np
import datetime 

def zone_loads(building,date,profile,weather,T_zone,m_v_zone,dt,frost):
    m_v_air = psychometric({'P':weather['pres_pa'],'Tdp':weather['t_dewp']},'w')
    n_z = len(building.zones['name'])
    T_air = [weather['t_dryb']*building.ctf['z_height'][i] -0.0065*building.ctf['z_height'][i] for i in range(n_z)] #outdoor air temp next to each zone
    T_mains = water_mains(building.water_main_temperature,date)
    
    ## Occupancy and latent heat from occupants
    loads = {}
    cat = ['convected','radiant','latent','visible','lost']
    for x in cat:
        loads[x] = [0 for i in range(n_z)]

    occupancy = [0 for i in range(n_z)]
    loads['occupancy'] = [0 for i in range(n_z)]
    loads['internal_gain'] = [0 for i in range(n_z)]
    for i in range(len(building.occupancy['zone'])):
        z_i = building.occupancy['zone'][i]
        sched_i = building.occupancy['schedule'][i]
        new_occ = profile[sched_i]*building.occupancy['nominal'][i]*building.zones['floor_area'][z_i]
        occupancy[z_i] += new_occ
        try:
            metabolic_rate = profile[building.occupancy['activity'][i]] # W/person
        except KeyError:
            metabolic_rate = 120  #120W/person
        sensible_heat = (6.461927 + 0.946892*metabolic_rate + .0000255737*metabolic_rate**2 + 7.139322*T_zone[z_i] -
            0.0627909*T_zone[z_i]*metabolic_rate + 0.0000589172*T_zone[z_i]*metabolic_rate**2 - 0.198550*T_zone[z_i]**2 + 
            0.000940018*T_zone[z_i]**2*metabolic_rate - 0.00000149532*T_zone[z_i]**2*metabolic_rate**2)

        loads['radiant'][z_i] += building.occupancy['radiant'][i]*new_occ*sensible_heat  #radiant is a portion of the sensible heat
        loads['convected'][z_i] += building.occupancy['convected'][i]*new_occ*sensible_heat
        loads['latent'][z_i] += new_occ*(metabolic_rate - sensible_heat)
        loads['occupancy'][z_i] += new_occ*metabolic_rate
    loads['internal_gain'] = [loads['internal_gain'][i] + loads['occupancy'][i] for i in range(n_z)]

    loads['multiplier'] = building.zones['multiplier']
    loads['area'] = building.zones['floor_area']

    ## Infiltration and Mixing
    infiltration = [0 for i in range(n_z)] #volumetric flows (m3/s)
    for j in range(len(building.infiltration['nominal'])):
        infiltration[building.infiltration['zone'][j]] = building.infiltration['nominal'][j]*profile[building.infiltration['schedule'][j]] #volumetric flows (m3/s)
    mixing = [[0 for i in range(n_z)] for j in range(n_z)] #volumetric flows (m3/s)
    for j in range(len(building.mixing['nominal'])):
        if building.mixing['type'][j]=='flow/person':
            mixing[building.mixing['receiving_zone'][j]][building.mixing['source_zone'][j]] = building.mixing['nominal'][j]*occupancy[building.mixing['source_zone'][j]]*profile[building.mixing['schedule'][j]] #volumetric flows (m3/s)
        else:
            mixing[building.mixing['receiving_zone'][j]][building.mixing['source_zone'][j]] = building.mixing['nominal'][j]*profile[building.mixing['schedule'][j]]#volumetric flows (m3/s)
    
    ## Equipment and Lighting
    prop = ['lighting_internal','plug_load','gas_load']
    for j in prop:
        loads[j] = [0 for i in range(n_z)]
        for i in range(len(getattr(building,j)['name'])):
            z_i = getattr(building,j)['zone'][i]
            sched_i = getattr(building,j)['schedule'][i]
            loads[j][z_i] += profile[sched_i]*getattr(building,j)['nominal'][i]*building.zones['floor_area'][z_i] 
            loads['internal_gain'][z_i] += loads[j][z_i]
            for c in cat:
                if c in getattr(building,j):
                    loads[c][z_i] += loads[j][z_i]*getattr(building,j)[c][i]
    ## Solar
    sunrise, sunset, azimuth, zenith = solar_calc(building.location['longitude'][0],building.location['latitude'][0],building.location['timezone'][0],[date])
    vect_2_sun = [float(np.sin(np.radians(zenith[0]))*np.sin(np.radians(azimuth[0]))),
                  float(np.sin(np.radians(zenith[0]))*np.cos(np.radians(azimuth[0]))),
                  float(np.cos(np.radians(zenith[0])))]## normal vector pointed at sun from azimuth (compass heading of sun) and zenith (angle beween sun and normal to surface of earth)
    #solar gain per unit area of exterior        
    if (weather['dir_norm_irr']+weather['dif_horz_irr'])>0:
        ext_gain = []
        for i in range(len(building.surfaces['exterior']['cos_phi'])):
            if building.surfaces['exterior']['boundary'][i]=='ground':
                ext_gain.append(0)
            else:    
                F_s = 0.5*(1+building.surfaces['exterior']['cos_phi'][i])
                F_g = 0.5*(1-building.surfaces['exterior']['cos_phi'][i])
                ext_gain.append(weather['dif_horz_irr']*(building.surfaces['exterior']['solar_absorptance'][i]*F_s) + building.site['ground_reflect']*weather['dif_horz_irr']*(building.surfaces['exterior']['solar_absorptance'][i]*F_g))
                cos_theta = vect_2_sun[0]*building.surfaces['exterior']['normal'][i][0] + vect_2_sun[1]*building.surfaces['exterior']['normal'][i][1] + vect_2_sun[2]*building.surfaces['exterior']['normal'][i][2] #dot product of two vectors
                SF = 1 #shading factor, needs development, pg 211
                if cos_theta>0:
                    ext_gain[i] += weather['dir_norm_irr']*building.surfaces['exterior']['solar_absorptance'][i]*SF*cos_theta #Absorbed direct normal radiation
    else:
        ext_gain = [0 for i in range(len(building.surfaces['exterior']['cos_phi']))]  
    
    ## Windows
    visible = [loads['visible'][building.windows['zone_index'][i]]/building.zones['interior_area'][building.windows['zone_index'][i]] for i in range(len(building.windows['area']))]
    long_wave = [loads['radiant'][building.windows['zone_index'][i]]/building.zones['interior_area'][building.windows['zone_index'][i]] for i in range(len(building.windows['area']))]
    window_gain,S = window_solar_transmittance(building.windows,building.surfaces,vect_2_sun,weather['dir_norm_irr'],weather['dif_horz_irr'],visible,long_wave)
    ## Refrigeration
    loads['case'],ref_gain,frost = refrigerated_cases(building,profile,T_zone,m_v_zone,frost,dt)
    loads['rack'],rack_gain = compressor_racks(building,loads['case'],T_air[0],T_zone,m_v_air)

    ## Water Equipment
    water_gain, water_heat = water_equipment_gains(building,profile,T_zone,m_v_zone,T_mains,dt)

    ## Exterior loads
    frac_dark = astronomical_clock_controlled_exterior_lights(date,dt,sunrise,sunset)
    loads['exterior'] = {}
    loads['exterior']['lighting'] = 0
    loads['exterior']['equipment'] = 0
    for j in building.exterior:
        prof = profile[building.exterior[j]['schedule']]
        if building.exterior[j]['type']=='lights':
            loads['exterior']['lighting'] += prof*building.exterior[j]['nominal']*frac_dark
        else:
            loads['exterior']['equipment'] += prof*building.exterior[j]['nominal']

    # group terms
    loads['latent'] = [loads['latent'][i] + water_gain['latent'][i] for i in range(n_z)]
    gains = {}
    gains['zone_sensible'] = [loads['convected'][i] + rack_gain['zone'][i] + ref_gain['zone_sensible'][i] + water_gain['sensible'][i] for i in range(n_z)]
    gains['zone_latent'] = [loads['latent'][i] + ref_gain['zone_latent'][i] for i in range(n_z)]
    gains['interior_surface'] = [(loads['radiant'][building.surfaces['zone_index'][i]] + loads['visible'][building.surfaces['zone_index'][i]])*building.surfaces['area'][i]/building.zones['interior_area'][building.surfaces['zone_index'][i]] for i in range(len(building.surfaces['area']))]
    gains['exterior_surface'] = ext_gain
    gains['hvac_sensible'] = [rack_gain['hvac'][i] + ref_gain['hvac_sensible'][i] for i in range(n_z)]
    gains['hvac_latent'] = ref_gain['hvac_latent']
    gains['window'] = window_gain
    gains['ref_rack_water_sensible'] = [rack_gain['zone'][i] + ref_gain['zone_sensible'][i] + water_gain['sensible'][i] for i in range(n_z)]
    gains['ref_rack_water_latent'] = [water_gain['latent'][i] + ref_gain['zone_latent'][i] for i in range(n_z)]

    return loads,gains,occupancy,mixing,infiltration,S,frost,T_mains,water_heat

def refrigerated_cases(building,profile,T_zone,m_v_zone,frost,dt):
    ##See page 1343 of the reference manual
    n_c = len(building.cases)
    n_z = len(building.zones['name'])
    loads = {}
    loads['electric'] = [0 for i in range(n_c)]
    loads['evap'] = [0 for i in range(n_c)]
    loads['case'] = []
    ref_gain = {}
    ref_gain['zone_sensible'] = [0 for i in range(n_z)] #cumulative sensible heat to all refrigerated cases from the building zone
    ref_gain['zone_latent'] = [0 for i in range(n_z)]
    ref_gain['hvac_sensible'] = [0 for i in range(n_z)] #cumulative sensible heat credits applied to HVAC
    ref_gain['hvac_latent'] = [0 for i in range(n_z)]
    j=-1
    for i in building.cases:
        j+=1
        bc = building.cases[i]
        loads['case'].append(i)
        z = building.zones['name'].index(bc['zone'])
        RAF = bc['return_air_frac']
        sched_i = bc['schedule']
        if bc['defrost_type']=='none':
                P_fan = bc['fan_power_per_length']*bc['length']
        elif bc['defrost_type']=='electric':
                sched_def = profile[bc['defrost_schedule']] #fraction of time defrosting
                sched_drip = profile[bc['defrost_drip_down_schedule']]
                P_def = bc['defrost_power_per_length']*bc['length']*sched_def
                latent_ratio = 1
                new_frost = bc['capacity_per_length']*bc['length']*bc['runtime_fraction']*bc['latent_heat_ratio']*latent_ratio*dt*(1-sched_drip)/(335000 + 2498000)
                max_melt_frost = P_def*dt/335000
                frost[j],Q_def = frost_calc(frost[j],new_frost,P_def,max_melt_frost,dt)
                P_fan = bc['fan_power_per_length']*bc['length']*(1-sched_def)
        elif bc['defrost_type']=='offcycle':
            #       sched_drip = profile[bc['defrost_drip_down_schedule']]
                P_fan = bc['fan_power_per_length']*bc['length']
        elif bc['defrost_type']=='hotgas':
            pass
        elif bc['defrost_type']=='hotbrine':
            pass
        RH = psychometric({'Tdb':T_zone[z],'w':m_v_zone[z]},'RH')
        Tdp = psychometric({'Tdb':bc['temperature'],'w':m_v_zone[z]},'Tdp')

        P_lights = bc['standard_lighting_per_unit_length']*bc['length']*profile[bc['light_schedule']]
        if bc['anti_sweat_control']=='none':
            P_as = 0
        elif bc['anti_sweat_control']=='constant':
            P_as = bc['anti_sweat_heater_per_length']*bc['length']
        elif bc['anti_sweat_control']=='RelativeHumid' or bc['anti_sweat_control']=='linear':
            P_as = bc['anti_sweat_heater_per_length']*bc['length']*(1 - (bc['rated_ambient_RH'] - RH)/(bc['rated_ambient_RH'] - bc['humidity_at_zero_percent']))
        elif bc['anti_sweat_control']=='dewpoint':
            pass
        elif bc['anti_sweat_control']=='heatbalance' or bc['anti_sweat_control']=='HeatBalanceMethod':
            ### first find R_case at nominal conditions
            Tdp_rated = psychometric({'Tdb':bc['rated_ambient_T'],'RH':bc['rated_ambient_RH']},'Tdp')
            R_case = (Tdp_rated - bc['temperature'])/(bc['anti_sweat_heater_per_length']/bc['height'] - (Tdp_rated - bc['rated_ambient_T'])/0.3169)
            P_as = ((Tdp - T_zone[z])*bc['height']/0.3169 + (Tdp - bc['temperature'])/R_case)*bc['length']
        P_as = max(P_as,bc['minimum_anti_sweat_per_length']*bc['length'])
        Q_restock = profile[bc['restock_schedule']]*bc['length']
        try: 
            credit_frac = profile[bc['case_credit_fraction_schedule']]
        except:
            credit_frac = 1
        Q_sensible_credit_rated = (bc['capacity_per_length']*bc['length']*bc['runtime_fraction']*(1-bc['latent_heat_ratio']) - P_lights*bc['light_to_case'] - P_as*bc['anti_sweat_heat_to_case'] - P_fan)
        Q_sensible_credit = Q_sensible_credit_rated*((T_zone[z] - bc['temperature'])/(bc['rated_ambient_T'] - bc['temperature']))*credit_frac
        ref_gain_sensible = (P_lights*(1-bc['light_to_case']) + P_as*(1-bc['anti_sweat_heat_to_case']) - Q_sensible_credit)*profile[sched_i]
        loads['evap'][j] = (Q_sensible_credit + P_fan + P_lights*bc['light_to_case'] + P_as*bc['anti_sweat_heat_to_case'] + Q_restock + Q_def)*profile[sched_i]
        ## Latent loads
        if bc['latent_credit_curve_type']=='casetemperaturemethod':
            latent_ratio = 1 - (bc['rated_ambient_RH'] - RH)*eval_curve(building.curves,bc['latent_credit_curve_name'],[bc['temperature']])
        elif bc['latent_credit_curve_type']=='relativehumiditymethod':
            latent_ratio = eval_curve(building.curves,bc['latent_credit_curve_name'],[RH])
        elif bc['latent_credit_curve_type']=='dewpointmethod':
            latent_ratio = eval_curve(building.curves,bc['latent_credit_curve_name'],[Tdp])
        else:
            print('check names of case latent credit curve')
        if latent_ratio>1:
            latent_ratio = 1
        elif latent_ratio<0:
            latent_ratio = 0
        ref_gain_latent = bc['capacity_per_length']*bc['length']*bc['latent_heat_ratio']*bc['runtime_fraction']*credit_frac*latent_ratio

        ##split into zone & hvac due to return_air fraction
        ref_gain['hvac_sensible'][z] += ref_gain_sensible*RAF
        ref_gain['zone_sensible'][z] += ref_gain_sensible*(1-RAF)
        ref_gain['hvac_latent'][z] += ref_gain_latent*RAF
        ref_gain['zone_latent'][z] += ref_gain_latent*(1-RAF)

        loads['electric'][j] = (P_fan + P_lights + P_as)*profile[sched_i]
        if bc['defrost_type']=='electric':
            loads['electric'][j] += P_def*profile[sched_i]
    return loads,ref_gain,frost

def compressor_racks(building,case_loads,T_air,T_zone,m_v_air):
    n_r = len(building.racks)
    loads = {}
    loads['electric'] = [0 for i in range(n_r)]
    loads['water'] = [0 for i in range(n_r)]
    loads['evap'] = [0 for i in range(n_r)]
    Q_reclaim = [0 for i in range(n_r)]
    rack_gain = {}
    n_z=len(building.zones['name'])
    rack_gain['zone'] = [0 for i in range(n_z)] 
    rack_gain['hvac'] = [0 for i in range(n_z)]
    j = -1
    for i in building.racks:
        j +=1
        br = building.racks[i]
        k = [case_loads['case'].index(br['cases'][x]) for x in range(len(br['cases']))]
        loads['evap'][j] = sum([case_loads['evap'][c] for c in k])
        if br['heat_reject_location']=='outdoors':
            z = -1
            if br['condensor_type']=='aircooled':
                T_eval = T_air
            else: # br['condensor_type']=='evapcooled':
                print ('error: non air cooled rack condensor' )
                # Twb = psychometric({'Tdb':T_air,'w':m_v_air},'Twb')
                # T_effective = Twb + (1-evap_condenser_effectiveness)*(T_air - Twb)
                # T_eval = T_effective
                T_eval = T_air
        else:
            z = building.zones['name'].index(br['heat_reject_location'])
            T_eval = T_zone[z]

        COP = br['design_COP']*eval_curve(building.curves,br['COP_curve'],[T_eval])
        if br['condensor_type']=='watercooled':
            fan_power = 0
            loads['water'][j] = loads['evap'][j]/COP
        else:
            try:
                fan_power = br['fan_power']*eval_curve(building.curves,br['fan_power_temperature_curve'],[T_eval])
            except:
                fan_power = br['fan_power']

        if br['heat_reject_location']!='outdoors' or br['condensor_type']=='aircooled' or br['condensor_type']=='evapcooled':
            loads['electric'][j] = loads['evap'][j]/COP + fan_power

        if z>=0:
            rack_gain['zone'][z] = sum(case_loads['evap'][k]*(1-building.cases['return_air_frac'][br['cases']]))/sum(case_loads['evap'][k])*(loads['evap'][j]/COP + fan_power)
            rack_gain['hvac'][z] = (loads['evap'][j]/COP + fan_power) - rack_gain['zone'][z]
        else:
            Q_reclaim[j] = 0.3*loads['evap'][j]*(1 + 1/COP) #possible to reclaim up to 30% waste heat
    return loads,rack_gain

def astronomical_clock_controlled_exterior_lights(date,dt,sunrise,sunset):
    shift = .04 #.125 % fraction of day exterior lights remain on after sunrise
    if date.hour<(sunrise*24+shift):
        frac_dark = 1
    elif date.hour-dt<=(sunrise*24+shift):
        frac_dark = 1-((sunrise*24+shift) - (date.hour-dt))/dt
    elif date.hour<(sunset*24-shift):
        frac_dark = 0
    elif (date.hour-dt)<(sunset*24-shift):
        frac_dark =((sunset*24-shift) - (date.hour-dt))/dt
    else:
        frac_dark = 1
    return frac_dark

def frost_calc(frost,new_frost,P_def,max_melt_frost,dt):
    frost = frost + new_frost
    Q_def = 0
    if P_def>0:
        if frost>max_melt_frost:
            frost = frost-max_melt_frost
        else:
            frost = 0
            Q_def = P_def - frost*335000/dt
    return frost,Q_def

def water_mains(main,date):
    julian_day = round((date - datetime.datetime(date.year,1,1)).seconds/(3600*24) +.5)
    T_out_avg = main['outdoor_average'][0]*9/5+32 #convert to F 
    T_max_diff = main['max_temp_difference'][0]*9/5 #convert to delta F
    ratio = 0.4 + 0.01*(T_out_avg-44)
    lag = 35 - 1.0*(T_out_avg-44)
    T_mains = float((T_out_avg+6) + ratio*(T_max_diff/2)*np.sin((0.986*(julian_day - 15 - lag) -90)*np.pi/180))
    T_mains = (T_mains-32)*5/9 #convert back to C
    return T_mains

def water_equipment_gains(building,profile,T_zone,m_v_zone,T_mains,dt):
    Cp_water = 4186 #J/kg*K
    den_water = 995 #kg/m^3
    n_z = len(building.zones['name'])
    water_gain = {}
    water_gain['sensible'] = [0 for i in range(n_z)]
    water_gain['latent'] = [0 for i in range(n_z)]
    water_heat = 0
    for j in building.water_use:
        z = building.zones['name'].index(building.water_use[j]['zone'])
        _,_,T_target,T_hot,T_cold = water_equipment(building.water_use[j],profile,T_mains,None)
        ##find gain into zones
        
        m_mix = building.water_use[j]['peak_flow']*profile[building.water_use[j]['flow_schedule']]*den_water 
        m_hot = m_mix*(T_target- T_cold)/(T_hot - T_cold)
        Q_total = m_mix*Cp_water*(T_target - T_zone[z]) #Watts of energy relative to zone     
        e_num = building.plant_demand_equip['name'].index(j)
        loop_connect = building.plant_demand_equip['loop'][e_num]
        if building.plant_loop['type'][loop_connect]!='Heating':
            water_heat += (m_hot*Cp_water*(T_hot - T_cold))*building.zones['multiplier'][z] #only if not on a plant loop
        water_gain['latent'][z] = profile[building.water_use[j]['latent_frac_schedule']]*Q_total
        water_gain['sensible'][z] = Q_total*profile[building.water_use[j]['sensible_frac_schedule']] #energy added to zone (without zone multiplier)
    return water_gain, water_heat


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eagers',
 'eagers.basic',
 'eagers.building',
 'eagers.building.basic',
 'eagers.building.components',
 'eagers.building.process',
 'eagers.class_definition',
 'eagers.config',
 'eagers.forecasting',
 'eagers.plot',
 'eagers.read',
 'eagers.setup',
 'eagers.simulate',
 'eagers.solver',
 'eagers.update',
 'eagers.write']

package_data = \
{'': ['*']}

install_requires = \
['ecos>=2.0.7,<3.0.0',
 'matplotlib>=3.4.2,<4.0.0',
 'numpy>=1.21.1,<2.0.0',
 'openpyxl>=3.0.7,<4.0.0',
 'scipy>=1.7.0,<2.0.0',
 'tables>=3.6.1,<4.0.0']

setup_kwargs = {
    'name': 'eagers',
    'version': '0.1.3',
    'description': 'Efficient Allocation of Grid Energy Resources including Storage',
    'long_description': None,
    'author': 'Dustin McLarty',
    'author_email': 'dustin.mclarty@wsu.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.9',
}


setup(**setup_kwargs)



'''
 http://www.ce.utexas.edu/prof/Novoselac/classes/ARE383/Handouts/F01_06SI.pdf
Nomenclature
 Tdb = dry bulb temperature in Celcius
 Twb = wet bulb temperature in Celcius
 Tdp = dew point temperature in Celcius
 w = moisture content (kg H20 / kg dry air)
 h enthalpy J/kg*K
 RH relative humidity
 P pressure in Pa
'''
import numpy as np
import copy

def psychometric(val,prop):
    if prop in val:
        del val[prop]
    kys = list(val.keys())
    try:
        n = len(val[kys[0]])
        air = copy.deepcopy(val)
        single = False
    except TypeError:
        air = {}
        for i in kys:
            air[i] = [copy.deepcopy(val[i])]
        single = True
        n=1
    if 'P' not in air:
        air['P'] = [101325 for i in range(n)]# atmospheric pressure (Pa)
        # air['P'] = 101325*(1-2.25577e-5*z)**5.2559 # atmospheric pressure function of height z(Pa)
        # air['Tdb'] = 15-0.0065*z # atmospheric  temperature function of height z(C)
    if 'h' in air and 'w' in air:
        air['Tdb'] = [(air['h'][i]/1000 - air['w'][i]*2501)/(1.006 + air['w'][i]*1.86) for i in range(n)] #Temperature in C and converting enthalpy to kJ/kg and humidity in massH2O/massAir
    elif 'h' in air and 'Tdb' in air:
        air['w'] = [(air['h'][i]/1000 - 1.006*air['Tdb'][i])/(2501 + 1.86*air['Tdb'][i]) for i in range(n)] #kJ/kg 
    elif 'RH' in air and 'Tdb' in air:
        satP = sat_press(air['Tdb'])
        P_H2O = [air['RH'][i]/100*satP[i]  for i in range(n)] #kPa
        air['w'] = [.621945*(P_H2O[i]/(air['P'][i]-P_H2O[i])) for i in range(n)]
    elif 'Twb' in air and 'w' in air:
        satP = sat_press(air['Twb']) #very small error introduced using the wet_bulb temperature for saturation pressure
        W_s = [0.62198*satP[i]/(air['P'][i]-satP[i]) for i in range(n)]
        air['Tdb'] = [(air['Twb'][i]*(4.186*air['w'][i]-2.381*W_s[i] + 1.006) +2501*W_s[i] - air['w'][i]*2501)/(1.805*air['w'][i]+1.006) for i in range(n)]
    elif 'Twb' in air:
        # assumes saturated
        air['Tdb'] = copy.deepcopy(air['Twb'])
        satP = sat_press(air['Tdb'])
        air['w'] = [0.62198*satP[i]/(air['P'][i]-satP[i]) for i in range(n)]
    elif 'Tdp' in air:
        # assumes saturated
        air['Tdb'] = copy.deepcopy(air['Tdp'])
        satP = sat_press(air['Tdp'])
        air['w'] = [0.62198*satP[i]/(air['P'][i]-satP[i]) for i in range(n)]
    elif 'h' in air:
        # assumes saturated
        air['Tdb'] = [10 for i in range(n)]
        for i in range(5):
            satP = sat_press(air['Tdb'])
            air['w'] = [0.62198*satP[i]/(air['P'][i]-satP[i]) for i in range(n)]
            air['Tdb'] = [0.7*air['Tdb'][i]  + 0.3*(air['h'][i]/1000 - air['w'][i]*2501)/(1.006 + air['w'][i]*1.86) for i in range(n)]
    
    
    if prop =='h':
            air['h'] = [1000*(1.006*air['Tdb'][i] + air['w'][i]*(2501 + 1.86*air['Tdb'][i])) for i in range(n)]
    elif prop =='Twb':
            #not perfect but close (0.26C) % https://journals.ametsoc.org/doi/full/10.1175/JAMC-D-11-0143.1
            satP = sat_press(air['Tdb'])
            P_H2O = [air['P'][i]*air['w'][i]/(0.62198 + air['w'][i]) for i in range(n)]
            air['RH'] = [P_H2O[i]/satP[i]*100  for i in range(n)] #Relative humidity in percent
            air['Twb'] = [float(air['Tdb'][i]*np.arctan(0.151977*(air['RH'][i] + 8.313659)**.5) + np.arctan(air['Tdb'][i] + air['RH'][i]) - np.arctan(air['RH'][i] - 1.676331) + 0.00391838*(air['RH'][i])**(3/2)*np.arctan(0.023101*air['RH'][i]) - 4.686035)  for i in range(n)]
    elif prop =='Tdp':
            air['Tdp'] = dew_point(air['Tdb'],air['P'],air['w'])
    elif prop =='RH':
            satP = sat_press(air['Tdb'])
            P_H2O = [air['P'][i]*air['w'][i]/(0.62198 + air['w'][i]) for i in range(n)]
            air['RH'] = [P_H2O[i]/satP[i]*100 for i in range(n)] #Relative humidity in percent
    elif prop =='T':
            air['T'] = copy.deepcopy(air['Tdb'])
    if single:
        kys = list(air.keys())
        for i in kys:
            air[i] = copy.deepcopy(air[i][0])
    out = copy.deepcopy(air[prop])
    return out

def dew_point(T,P,w):
    n = len(T)
    P_H2O = [P[i]*w[i]/(0.62198 + w[i])/1000 for i in range(n)] # kPa 0.62198 is ratio of molar masses
    Tdp = [[] for i in range(n)]
    for i in range(n):
        if T[i]<0:
            Tdp[i] = float(6.09 + 12.608*np.log(P_H2O[i]) + 0.4959*np.log(P_H2O[i])**2)
        else:
            Tdp[i] = float(6.54 + 14.526*np.log(P_H2O[i]) + 0.7389*np.log(P_H2O[i])**2 + 0.09486*np.log(P_H2O[i])**3 + 0.4569*(P_H2O[i])**0.1984) #Dew point from partial pressure of water using ASHRAE 2013 Fundamentals eqn 39 valid from 0C to 93C
    return Tdp

def sat_press(T):
    n = len(T)
    Tdb_K = [T[i]+273.15 for i in range(n)] #Tdb (Kelvin)
    satP = [[] for i in range(n)]
    for i in range(n):
        if T[i]<0:
            satP[i] = float(np.exp((-5.6745359e3)/Tdb_K[i] + 6.3925247 - 9.677843e-3*Tdb_K[i] + 6.22157e-7*Tdb_K[i]**2 + 2.0747825e-9*Tdb_K[i]**3 -9.484024e-13*Tdb_K[i]**4 + 4.1635019*np.log(Tdb_K[i]))) # Pa, saturated water vapor pressure ASHRAE 2013 fundamentals eq. 6 in kPa valid for 0 to 200C
        else:
            satP[i] = float(np.exp((-5.8002206e3)/Tdb_K[i] + 1.3914993 - 4.8640239e-2*Tdb_K[i] + 4.1764768e-5*Tdb_K[i]**2 - 1.4452093e-8*Tdb_K[i]**3 + 6.5459673*np.log(Tdb_K[i]))) # Pa saturated water vapor pressure ASHRAE 2013 fundamentals eq. 6 in kPa valid for 0 to 200C
    return satP


'''
Run simulation of multi-zone building
Inputs are the parameters of the building in 'building'
The initial state of zone, surface temperatures and surface humidity
The weather and dates of the simulation period, and if it is a design day
'''

from eagers.building.basic.leward_calc import leward_calc
from eagers.building.basic.wind_speed_calc import wind_speed_calc
from eagers.building.basic.thermostat_humidistat import thermostat_humidistat
from eagers.building.process.load_sched import load_sched
from eagers.building.process.initial_param import initial_param
from eagers.building.process.reset_loop_param import reset_loop_param
from eagers.building.process.update_model_states_implicit import update_model_states_implicit
from eagers.building.process.zone_loads import zone_loads
from eagers.building.process.ideal_hvac import ideal_hvac
from eagers.building.process.sim_air_loops import sim_air_loops
from eagers.building.process.manage_plant_equip import manage_plant_equip
from eagers.building.components.unitary_systems import unitary_systems
from eagers.building.components.air_terminals import air_terminals

def run_building(building,T_zone,T_surf,humidity,weather,date,des_day):
    air_density = 1.204  #kg/m^3 %standard air density (1.204 kg/m3) adjusted for the local barometric pressure (standard barometric pressure corrected for altitude, ASHRAE 1997 HOF pg. 6.1).
    nz = len(building.zones['name'])
    Q_transfer = [] #energy transfer to zone
    m_sys = [] #air mass flow to zone
    frost = [[0 for i in range(len(building.cases))]] #need to make as initial condition that is passed from previous time step
    cat = ['net_elec','heat_elec','heat_gas','cool_elec','fan_elec','other_fans','water_gas','water_elec','tower_elec','pumps']
    e_use={}
    facility = {}
    Q_test_plots = {}
    weather_now = {}
    for i in cat:
        e_use[i] = 0
        facility[i] = []
    facility['heating_load'] = []
    facility['cooling_load'] = []
    facility['tower_load'] = []
    facility['supply'] = []
    facility['return_'] = []
    T,air_nodes,plant_nodes,T_supply0,T_return0 = initial_param(building,T_zone[0],T_surf[0])
    wh_tank = {}
    for i in building.water_heater:
        wh_tank[i] = {}
        wh_tank[i]['T'] = load_sched(building.schedule,building.holidays,[date[0]],[building.water_heater[i]['temperature_schedule']],des_day)
        wh_tank[i]['on'] = False
    w = {}
    w['zone'] = humidity[0]
    t_last = date[0]
    for t in range(len(date)-1):
        dt = (date[t+1] - t_last).seconds
        if dt <0:
            dt = dt + 24*3600 #necessary for warm-up period when date jumps back to start of day
        for i in weather:
            weather_now[i] = weather[i][t]
        leward = leward_calc(weather_now['wdir'],building.surfaces['exterior']['normal'])
        wind_speed = wind_speed_calc(weather_now['wspd'],building.ctf['s_height'],building.site['terrain'])
        
        T,w,air_nodes,plant_nodes = reset_loop_param(building,T,w,air_nodes,plant_nodes,weather_now,date[t+1])
        schedules = load_sched(building.schedule,building.holidays,[date[t+1]],None,des_day)

        ###Load values specific to this moment in time
        T_set,w_set = thermostat_humidistat(building,schedules,T['zone']) ## HVAC temperature Schedules
        loads,gains,occupancy,mixing,infiltration,S,frost_t,T_mains,water_heat = zone_loads(building,date[t+1],schedules,weather_now,T['zone'],w['zone'],dt,frost[t])
        
        ###Determine HVAC setpoints
        air_nodes,central_flow,direct_flow,mixed_air_loop,direct_return_air,plenum,T,w,fan_avail = ideal_hvac(building,schedules,leward,wind_speed,S,air_nodes,weather_now['t_dryb'],mixing,infiltration,gains,T,w,T_set,w_set,occupancy,dt,des_day)
        if not des_day:
            ###Compute performance of HVAC system 
            for iter in range(3): #iterate to ensure water loops reach correct temperature
                for x in cat:
                    e_use[x] = 0
                plant_nodes['demand_temperature'] = [plant_nodes['demand_temperature'][building.plant_demand_nodes['inlet_node'][building.plant_demand_nodes['loop'][i]]] for i in range(len(building.plant_demand_nodes['loop']))]
                supply_flow,e_use,plant_nodes = sim_air_loops(building,air_nodes,mixed_air_loop,central_flow,w['loop'],weather_now['t_dryb'],w['air'],plant_nodes,e_use,fan_avail)
                supply_flow,e_use,plant_nodes = air_terminals(building,air_nodes,supply_flow,e_use,plant_nodes,T['central'])
                supply_flow,plant_nodes,e_use = unitary_systems(building,air_nodes,supply_flow,direct_flow,direct_return_air,T['direct'],weather_now['t_dryb'],w['air'],fan_avail,plant_nodes,e_use)#zone equipment not part of central air loops
                plant_nodes,e_use,wh_tank,loop_load = manage_plant_equip(building,schedules,plant_nodes,e_use,dt,weather_now['t_dryb'],w['air'],T_mains,wh_tank,T_supply0,T_return0)
        
        e_use['water_gas'] += water_heat/building.impact_factor['steam_efficiency'][0] ##unconnected water heat is considered purchased heat
        nonHVAC_electric = sum([loads['lighting_internal'][z]*loads['multiplier'][z] for z in range(nz)]) + sum([loads['plug_load'][z]*loads['multiplier'][z] for z in range(nz)]) + loads['exterior']['lighting'] + loads['exterior']['equipment'] + sum(loads['case']['electric']) + sum(loads['rack']['electric'])
        e_use['net_elec'] = (nonHVAC_electric + e_use['pumps'] + e_use['heat_elec'] + e_use['cool_elec'] + e_use['fan_elec'] + e_use['water_elec'] + e_use['tower_elec'])

        ## simulate a time step
        T['zone'],T['surf'],w['zone'],e_dict = update_model_states_implicit(building,T,S,wind_speed,leward,gains,dt,w,mixing,plenum,infiltration,supply_flow)
        if any([not building.zones['air_loop'][j] is None and T['zone'][j]<T_set['heat'][j]- building.hvac['managers']['thermostat_tolerance'][building.zones['air_loop'][j]] and not T_set['no_hvac'][j] for j in range(len(T['zone_est']))]):
            print('investigate why building not reaching setpoint')
        frost.append(frost_t)
        ## update logging variables and move to next time step
        T_zone.append(T['zone'])
        T_surf.append(T['surf'])
        m_sys.append([air_density*(central_flow[i] + direct_flow[i]) for i in range(nz)])
        humidity.append(w['zone'])
        Q_transfer.append([air_density*(central_flow[i]*(1006 + 1860*w['central'][i])*(T['central'][i] - T['zone'][i])
                         + direct_flow[i]*(1006 + 1860*w['zone'][i])*(T['direct'][i] - T['zone'][i]) 
                         + (1006 + 1860*w['zone'][i])*(sum([plenum[i][j]*T['zone'][j] for j in range(nz)]) 
                         - sum(plenum[i])*T['zone'][i]))*building.zones['multiplier'][i] for i in range(nz)]) #Sensible energy transfer to the zone.
        for l in range(len(building.plant_loop['name'])):
            T_supply0[l] = plant_nodes['demand_temperature'][building.plant_demand_nodes['inlet_node'][l]] #half loop tank temperature from previous time step (pg 468)
            T_return0[l] = plant_nodes['supply_temperature'][building.plant_supply_nodes['inlet_node'][l]] #half loop tank temperature from previous time step (pg 468)
        for i in cat:
            facility[i].append(e_use[i])
        facility['heating_load'].append(sum(loop_load[0]))
        facility['cooling_load'].append(sum(loop_load[1]))
        facility['tower_load'].append(sum(loop_load[2]))
        facility['supply'].append(T_supply0)
        facility['return_'].append(T_return0)
        t_last = date[t+1]
    return T_zone,T_surf,humidity,frost,m_sys,Q_transfer,facility,Q_test_plots



from eagers.building.process.buildings_step import optim_T_set
from eagers.building.basic.leward_calc import leward_calc
from eagers.building.basic.wind_speed_calc import wind_speed_calc
from eagers.building.basic.thermostat_humidistat import thermostat_humidistat
from eagers.building.basic.area_weighted_temperature import area_weighted_temperature
from eagers.building.process.load_sched import load_sched
from eagers.building.process.initial_param import initial_param
from eagers.building.process.reset_loop_param import reset_loop_param
from eagers.building.process.update_model_states_implicit import update_model_states_implicit
from eagers.building.process.zone_loads import zone_loads
from eagers.building.process.ideal_hvac import ideal_hvac
from eagers.building.process.sim_air_loops import sim_air_loops
from eagers.building.process.net_water_loop_supply import net_water_loop_supply
from eagers.building.components.unitary_systems import unitary_systems
from eagers.building.components.air_terminals import air_terminals

def building_response(buildings,observer,weather_now,actual_build,dt,setpoint,date):
    #use the temperature setpoints from the optimization to find the actual zone temperatures
    n_b = len(buildings)
    net_electric = [0 for i in range(n_b)]
    cat = ['heat_elec','heat_gas','cool_elec','fan_elec','water_gas','water_elec','tower_elec','pumps','other_fans',]
    e_use = {}
    w = {}
    for i in range(len(cat)):
        e_use[cat[i]] = 0
    for k in range(n_b):
        building = buildings[k]
        heating = setpoint['district_heat'][k]
        cooling = setpoint['district_cool'][k]
        T_optim = setpoint['temperature'][k]
        nz = len(building.zones['name'])
        w['zone'] = observer['building_humidity'][k]
        frost = observer['building_frost'][k]
        T,air_nodes,plant_nodes,_,_= initial_param(building,observer['building_zone_temp'][k],observer['building_surf_temp'][k])
        for l in range(len(building.plant_loop['name'])):
            plant_nodes['demand_temperature'][building.plant_demand_nodes['loop'] == l] = observer['building_supply'][k][l]
            plant_nodes['supply_temperature'][building.plant_supply_nodes['inlet_node'][l]] = observer['building_return'][k][l]
        leward = leward_calc(weather_now['wdir'],building.surfaces['exterior']['normal'])
        wind_speed = wind_speed_calc(weather_now['wspd'],building.ctf['s_height'],building.site['terrain'])
        
        T,w,air_nodes,plant_nodes = reset_loop_param(building,T,w,air_nodes,plant_nodes,weather_now,observer['building_timestamp'][k])

        #parameters that could be overridden by 'actual' measurements
        if 'schedules' in actual_build:
            schedules = actual_build['schedules'][k]
        else:
            schedules = load_sched(building.schedule,building.holidays,[date],None,False)
        T_set,w_set = thermostat_humidistat(building,schedules,T['zone']) ## HVAC temperature Schedules
        if 'T_set' in actual_build:
            T_set = actual_build['T_set'][k]
        if 'w_set' in actual_build:
            w_set = actual_build['w_set'][k]
        ## extrapolate 'optimal' building temp to all zones
        T_set_mod = optim_T_set(T_set,T_optim,actual_build['Tz_nominal'][k],building)
        
        ###Load values specific to this moment in time
        loads,gains,occupancy,mixing,infiltration,S,frost,T_mains,water_heat = zone_loads(building,date,schedules,weather_now,T['zone'],w['zone'],dt,frost)
        if 'loads' in actual_build:
            loads = actual_build['loads'][k]
        if 'gains' in actual_build:
            gains = actual_build['gains'][k]
        if 'occupancy' in actual_build:
            occupancy = actual_build['occupancy'][k]
        if 'mixing' in actual_build:
            mixing = actual_build['mixing'][k]
        if 'infiltration' in actual_build:
            infiltration = actual_build['infiltration'][k]
        if 'T_mains' in actual_build:
            T_mains = actual_build['T_mains'][k]
        if 'water_heat' in actual_build:
            water_heat = actual_build['water_heat'][k]
        if 'S' in actual_build:
            S = actual_build['S'][k]
        ###Determine HVAC setpoints
        air_nodes,central_flow,direct_flow,mixed_air_loop,direct_return_air,plenum,T,w,fan_avail = ideal_hvac(building,schedules,leward,wind_speed,S,air_nodes,weather_now['t_dryb'],mixing,infiltration,gains,T,w,T_set_mod,w_set,occupancy,dt,False)
        for iter in range(3): #iterate to ensure water loops reach correct temperature
            for x in cat:
                e_use[x] = 0
            plant_nodes['demand_temperature'] = [plant_nodes['demand_temperature'][building.plant_demand_nodes['inlet_node'][building.plant_demand_nodes['loop'][i]]] for i in range(len(building.plant_demand_nodes['loop']))]
            supply_flow,e_use,plant_nodes = sim_air_loops(building,air_nodes,mixed_air_loop,central_flow,w['loop'],weather_now['t_dryb'],w['air'],plant_nodes,e_use,fan_avail)
            supply_flow,e_use,plant_nodes = air_terminals(building,air_nodes,supply_flow,e_use,plant_nodes,T['central'])
            supply_flow,plant_nodes,e_use = unitary_systems(building,air_nodes,supply_flow,direct_flow,direct_return_air,T['direct'],weather_now['t_dryb'],w['air'],fan_avail,plant_nodes,e_use)#zone equipment not part of central air loops
            e_use,plant_nodes,loop_load = net_water_loop_supply(building,schedules,dt,plant_nodes,e_use,T_mains,weather_now['t_dryb'],w['air'],heating,cooling)
            
        e_use['water_gas'] += water_heat/building.impact_factor['steam_efficiency'][0] #unconnected water heat is considered purchased heat
        nonHVAC_electric = sum([loads['lighting_internal'][z]*loads['multiplier'][z] for z in range(nz)]) + sum([loads['plug_load'][z]*loads['multiplier'][z] for z in range(nz)]) + loads['exterior']['lighting'] + loads['exterior']['equipment'] + sum(loads['case']['electric']) + sum(loads['rack']['electric'])
        e_use['net_elec'] = nonHVAC_electric + e_use['pumps'] + e_use['heat_elec'] + e_use['fan_elec'] + e_use['water_elec'] + e_use['tower_elec']

        ## simulate a time step & update observer
        observer['building_zone_temp'][k],observer['building_surf_temp'][k],observer['building_humidity'][k],e_dict = update_model_states_implicit(building,T,S,wind_speed,leward,gains,dt,w,mixing,plenum,infiltration,supply_flow)
        for l in range(len(building.plant_loop['name'])):
            observer['building_supply'][k][l] = plant_nodes['demand_temperature'][building.plant_demand_nodes['inlet_node'][l]]
            observer['building_return'][k][l] = plant_nodes['supply_temperature'][building.plant_supply_nodes['inlet_node'][l]]
        A = observer['building_conditioned_floor_area'][k]
        observer['building_avg_temp'][k] =  sum([observer['building_zone_temp'][k][z]*A[z] for z in range(len(A))])/sum(A) #area_weighted_temperature %% %current average building zone temperature (used as IC in optimization)
        observer['building_timestamp'][k] = date #weather_now['timestamp']
        net_electric[k] = e_use['net_elec']/1000 #kW
    return observer,net_electric


"""Excel interface for hydro functionality.
"""
import datetime

from .excel_interface import BaseTemplateReader


class HydroTemplateReader(BaseTemplateReader):
    """Hydro template reader class."""

    _default_filename = 'hydro_template'
    _sheet_properties = {
        'Fossil': dict(
            read_method='_read_fossil_sheet',
        ),
        'Utilities': dict(
            read_method='_read_utilities_sheet',
        ),
        'Network': dict(
            read_method='_read_network_sheet',
        ),
        'Hydroelectric': dict(
            read_method='_read_hydroelectric_sheet',
        ),
        'Inflows': dict(
            read_method='_read_inflows_sheet',
        ),
        'Outflows': dict(
            read_method='_read_outflows_sheet',
        ),
    }

    @classmethod
    def _read_fossil_sheet(cls, ws):
        linelimits = 100
        
        header = [cell.value for cell in ws[1]]

        gens = []
           
        for row in ws.iter_rows(min_row=2, max_row=linelimits):
 
            gen = {}
            for key, cell in zip(header, row):
                gen[key] = cell.value
            gens.append(gen)

        return gens

    @classmethod
    def _read_utilities_sheet(cls, ws):
        linelimits = 100

        header = [cell.value for cell in ws[1]]

        gens = []

        for row in ws.iter_rows(min_row=2,max_row=linelimits):

            gen = {}
            for key, cell in zip(header, row):
                gen[key] = cell.value
                gens.append(gen)
        
        return gens

    @classmethod
    def _read_network_sheet(cls, ws):
        linelimits = 100
         
        header = [cell.value for cell in ws[1]]

        gens = []
        for row in ws.iter_rows(min_row=2,max_row=linelimits):
            gen = {}
            for key, cell in zip(header, row):
                gen[key] = cell.value
                gens.append(gen)

        return gens  

    @classmethod
    def _read_hydroelectric_sheet(cls, ws):
        linelimits = 100
         
        header = [cell.value for cell in ws[1]]

        gens = []
        for row in ws.iter_rows(min_row=2,max_row=linelimits):
            gen = {}
            for key, cell in zip(header, row):
                gen[key] = cell.value
                gens.append(gen)

        return gens

    @classmethod
    def _read_inflows_sheet(cls, ws):
        linelimits = 100
         
        header = [cell.value for cell in ws[1]]

        gens = []
        for row in ws.iter_rows(min_row=2,max_row=linelimits):
            gen = {}
            for key, cell in zip(header, row):
                gen[key] = cell.value
                gen = datetime.datetime.strptime(gen)
                gens.append(gen)

        return gens


    @classmethod
    def _read_outflows_sheet(cls, ws):
        linelimits = 100
         
        header = [cell.value for cell in ws[1]]

        gens = []
        for row in ws.iter_rows(min_row=2,max_row=linelimits):
            gen = {}
            for key, cell in zip(header, row):
                gen[key] = cell.value
                gen = datetime.datetime.strptime(gen)
                gens.append(gen)

        return gens
    
